/***************************************************************************
                          ai.cpp  -  description
                             -------------------
    begin                : Thu Jan 6 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: ai.c,v 1.3 2000/11/25 14:48:53 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ai.h"

aiStruct AI;
extern Image red_carImg[4];
extern carStruct blue_car;
extern mapStruct *map;

void createAI(aiStruct *ai, mapStruct *map) {
	int a;
	ai->number = map->redNum;
	ai->bot = (carStruct *)calloc(ai->number, sizeof(carStruct));
	ai->order = (int *)calloc(ai->number, sizeof(int));
	ai->data = (int *)calloc(ai->number, sizeof(int));
	// Create the bots
	for (a = 0; a < ai->number; a++) {
		ai->bot[a].x = toMap(map->red_startX[a]);
		ai->bot[a].y = toMap(map->red_startY[a]);
		ai->bot[a].direction = ai->bot[a].toGoDir = UP;
		ai->bot[a].vel = map->aiVel;
		ai->bot[a].img[UP] = &red_carImg[UP];
		ai->bot[a].img[RIGHT] = &red_carImg[RIGHT];
		ai->bot[a].img[DOWN] = &red_carImg[DOWN];
		ai->bot[a].img[LEFT] = &red_carImg[LEFT];
		ai->bot[a].ai = -1;
		ai->order[a] = CONFUSED;
		ai->data[a] = map->enemywait;
	}

}


void orderAI(aiStruct *ai, mapStruct *map) {
	int a, b;
	int min_dist = map->width + map->height;
	carStruct *closest = NULL;

	// See which car is closest of the blue guy
	for (a = 0; a < ai->number; a++) {
		// Only cars that would fit are included
		b = getDistance(toTile(ai->bot[a].x), toTile(ai->bot[a].y), toTile(blue_car.x), toTile(blue_car.y));
		if (b < min_dist) {
			closest = &ai->bot[a];
			min_dist = b;
		}
	}

	// The closest car follow the blue guy, the others, just wander around
	for (a = 0; a < ai->number; a++) {
		if (ai->order[a] == CONFUSED)
			continue;
		if (closest == &ai->bot[a]) {
			ai->order[a] = GET_BLUE;
		}
		else
			ai->order[a] = WANDER;
  }
}

void moveAI(aiStruct *ai, mapStruct *map) {
	int a;
	for (a = 0; a < ai->number; a++) {
		switch(ai->order[a]) {
			case WANDER:
				carAI(&ai->bot[a], map);
				move(&ai->bot[a], map);
				break;
			case GET_BLUE:
				goTo(&ai->bot[a], map, toTile(blue_car.x), toTile(blue_car.y));
				if (!move(&ai->bot[a], map)) {
					carAI(&ai->bot[a], map);
					move(&ai->bot[a], map);
				}
				break;
			case CONFUSED:
				ai->bot[a].img[ai->bot[a].direction] = &red_carImg[(ai->data[a] % 8) / 2];
				ai->data[a]--;
				break;
		}
	}
}

int getDistance(int fx, int fy, int dx, int dy) {
	return (abs(fx - dx) + abs(fy - dy));
}

void goTo(carStruct *car, mapStruct *map, int dx, int dy) {
	int x, y, dirX, dirY;
	x = toTile(car->x);
	y = toTile(car->y);
	dirX = dx - x;
	dirY = dy - y;
//	while (!success) {
		if (abs(dirX) > abs(dirY)) {
			if (dirX < 0) {
				changeDir(car, map, LEFT);
			}	else {
				changeDir(car, map, RIGHT);
			}
		} else {
			if (dirY < 0) {
				changeDir(car, map, UP);
			} else {
				changeDir(car, map, DOWN);
			}
		}
//	}
}

void destroyAI() {
	if (AI.bot == NULL || AI.order == NULL || AI.data == NULL) {
		hdbg("This shouldn't be hapenning... (on destroyAI)");
		return;
	}
	free(AI.bot);
	AI.bot = NULL;
	free(AI.order);
	AI.order = NULL;
	free(AI.data);
	AI.data = NULL;
}
