/***************************************************************************
                          effects.c -  Special graphics effects for XRally
                             -------------------
    begin                : Sat Jan 1 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: effects.c,v 1.1 2000/11/25 14:48:53 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "global.h"
#include "graphics.h"
#include "effects.h"

static int _lookup_blend_cache(blend_cache_t *, unsigned long *pixel);
static void _add_blend_cache(blend_cache_t *, unsigned long old_pix, unsigned long new_pix);

void init_blend_cache(blend_cache_t *_cache) {
	XColor temp;
	// Lookup the color pixel value (is always good)
	alloc_color(&_cache->color);
	// Init the first node (uses the value of pixel 0)
	_cache->node = (node_t *)malloc(sizeof(node_t) * 1);
	_cache->node->old_pix = 0;
	temp.pixel = 0;
	query_color(&temp);
	temp.red = temp.red*(100-_cache->factor)/100 + _cache->color.red*_cache->factor/100;
  temp.blue = temp.blue*(100-_cache->factor)/100+ _cache->color.blue*_cache->factor/100;
	temp.green = temp.green*(100-_cache->factor)/100 + _cache->color.green*_cache->factor/100;
	alloc_color(&temp);
	_cache->node->new_pix = temp.pixel;
	_cache->node->next = NULL;
}

void free_blend_cache(blend_cache_t *_cache) {
	node_t *anode, *bnode;
	if (!_cache->node) {
		hdbg("This cache wasn't initialized!");
		return;
	}
	anode = _cache->node;
	while (anode->next) {
		bnode = anode->next;
		free(anode);
		anode = bnode;
	}
	free(anode);
	_cache->node = NULL;
}

static int _lookup_blend_cache(blend_cache_t *_cache, unsigned long *_pixel) {
	node_t *node;
	node = _cache->node;
	if (!node) {
		hdbg("This cache wasn't initialized!");
		return -1;
	}
	while (node->next != NULL) {
		if (node->old_pix == *_pixel) {
			*_pixel = node->new_pix;
			return 0;
		}
		else
			node = node->next;
	}
	return 1;
}

static void _add_blend_cache(blend_cache_t *_cache, unsigned long old_pix, unsigned long new_pix) {
	node_t *node;
	node_t *t_node;
	node = _cache->node;
	if (!node) {
		hdbg("This cache wasn't initialized!");
		return;
	}
	while (node->next != NULL)
		node = node->next;
	t_node = (node_t *)malloc(sizeof(node_t));
	t_node->old_pix = old_pix;
	t_node->new_pix = new_pix;
	t_node->next = NULL;
	node->next = t_node;
}

XImage *blend_image(Pixmap pix, int x, int y, int w, int h, blend_cache_t *cache) {
	XImage *window;
	XColor color;
	node_t *node = cache->node;
	unsigned long old_pix;
	if (!node) {
		hdbg("This cache wasn't initialized!");
		return NULL;
	}
	ldbg("Starting to blend...");
	window = get_image(pix, x, y, w, h);
	//XInitImage(window);
	for (y = 0; y < window->height; y++) {
		for (x = 0; x < window->width; x++) {
  		color.pixel = get_pixel(window, x, y);
			if (_lookup_blend_cache(cache, &color.pixel) != 0) {
				old_pix = color.pixel;
				query_color(&color);
				// Blend is the amount of the blend color, not the old color
				color.red = color.red*(100-cache->factor)/100 + cache->color.red*cache->factor/100;
				color.blue = color.blue*(100-cache->factor)/100+ cache->color.blue*cache->factor/100;
				color.green = color.green*(100-cache->factor)/100 + cache->color.green*cache->factor/100;
				alloc_color(&color);
				_add_blend_cache(cache, old_pix, color.pixel);
			}
			put_pixel(window, x, y, color.pixel);
		}
	}
	return window;
}

