/***************************************************************************
                          game.c -  description
                             -------------------
    begin                : Thu Nov 23 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "game.h"
#include "effects.h"
#include "graphics.h"
#include "map.h"
#include "car.h"
#include <stdio.h>
#include "menu.h"
#ifdef USE_SOUND
#include "sound.h"
#endif

// Radar

radarStruct radar;

// Pixmaps
Pixmap *tile;
Image blue_carImg[4];
Image red_carImg[4];

// Blend cache
extern blend_cache_t blend_cache;

// Smoke
Image smoke_tile;
GC smokeGC;
extern smoke_struct smoke;

// Crash tile
Image crash_tile;

// Extern info
extern carStruct blue_car;
extern carStruct red_car;
extern mapStruct *map;
extern aiStruct AI;
extern levelStruct level;
extern int posX, posY;
extern bool UPDATE;
extern bool BYE;
extern bool ENDMAP;
extern int mapResult;
extern bool crashed;
extern aiStruct AI;
extern FILE *log;
extern char livesString[10];
extern char mapString[15];
extern char *hiString;
extern scoreStruct score;
extern int actualFuel;

// Important colors:
unsigned long blue;
unsigned long red;
unsigned long green;
unsigned long black;
unsigned long yellow;
unsigned long white;

// last direction
unsigned int last_direction = 0;

// Fuel
GC scoreGC;
Pixmap scoreBuffer;

// Size of map
int SPRITE_SIZE = 48;
int TILE_WIDTH = 10;
int TILE_HEIGHT = 10;

char scrStr[5];

int load_images(){
	int a = 0;
	int w = 0, h = 0;
	int rw = 0, rh = 0;
	Pixmap temp;
	Image flag;
	Image pit;

	ldbg("Level pics: %d \t PIXS: %d", level.pics, PIXS);
	free(tile);
	tile = (Pixmap *)calloc(level.pics, sizeof(Pixmap));
	
	// Go to images dir
	a += chdir("images");
	if (a < 0) { 
		hdbg("Can't go to the images directory\nCurrent directory: %s", getcwd(NULL, -1));
	}
	

	a += chdir(level.tileset);
	if (a < 0) {
		dbg("Tileset %s not found!\nLoading from the default", level.tileset);
		strcpy(level.tileset, "default");
		a = chdir(level.tileset);
	}

	// Load the basic tiles
	
	// The road one is very special!
	// The future size of the pixmaps is based on it
	a += load_pixmap(&tile[ROAD], &temp, "road.xpm", &rw, &rh);
	if (a != 0) {
		hdbg("Can't find even the road pixmap! What kind of tileset is that?");
		hdbg("Using the default tileset and ignoring the level %s tileset", level.tileset);
		chdir("../default");
		strcpy(level.tileset, "default");
		load_pixmap(&tile[ROAD], &temp, "road.xpm", &rw, &rh);
	}

	LOAD_GAME_PIXMAP(&tile[OUT], NULL, "out.xpm");
	LOAD_GAME_PIXMAP(&tile[BLOCK], NULL, "block.xpm");
	LOAD_GAME_PIXMAP(&flag.pix, &flag.mask, "flag.xpm");
	LOAD_GAME_PIXMAP(&pit.pix, &pit.mask, "pit.xpm");
	LOAD_GAME_PIXMAP(&smoke_tile.pix, &smoke_tile.mask, "smoke.xpm");
	LOAD_GAME_PIXMAP(&crash_tile.pix, &crash_tile.mask, "crash.xpm");
	
	
	// Load the blue car tiles
	LOAD_GAME_PIXMAP(&blue_carImg[UP].pix, &blue_carImg[UP].mask, "blue_carUP.xpm");
	LOAD_GAME_PIXMAP(&blue_carImg[RIGHT].pix, &blue_carImg[RIGHT].mask, "blue_carRIGHT.xpm");
	LOAD_GAME_PIXMAP(&blue_carImg[DOWN].pix, &blue_carImg[DOWN].mask, "blue_carDOWN.xpm");
	LOAD_GAME_PIXMAP(&blue_carImg[LEFT].pix, &blue_carImg[LEFT].mask, "blue_carLEFT.xpm");
	

	// Load the red car tiles
	LOAD_GAME_PIXMAP(&red_carImg[UP].pix, &red_carImg[UP].mask, "red_carUP.xpm");
	LOAD_GAME_PIXMAP(&red_carImg[RIGHT].pix, &red_carImg[RIGHT].mask, "red_carRIGHT.xpm");
	LOAD_GAME_PIXMAP(&red_carImg[DOWN].pix, &red_carImg[DOWN].mask, "red_carDOWN.xpm");
	LOAD_GAME_PIXMAP(&red_carImg[LEFT].pix, &red_carImg[LEFT].mask, "red_carLEFT.xpm");
	

	// Load the custom pixmaps

	switch(level.pics) {
		case (15):
			LOAD_GAME_PIXMAP(&tile[14], NULL, "block6.xpm");
		case (14):
			LOAD_GAME_PIXMAP(&tile[13], NULL, "block5.xpm");
		case (13):
      LOAD_GAME_PIXMAP(&tile[12], NULL, "block4.xpm");
		case (12):
      LOAD_GAME_PIXMAP(&tile[11], NULL, "block3.xpm");
    case (11):
      LOAD_GAME_PIXMAP(&tile[10], NULL, "block2.xpm");
    case (10):
      LOAD_GAME_PIXMAP(&tile[9], NULL, "road6.xpm");
    case (9):
      LOAD_GAME_PIXMAP(&tile[8], NULL, "road5.xpm");
    case (8):
      LOAD_GAME_PIXMAP(&tile[7], NULL, "road4.xpm");
    case (7):
      LOAD_GAME_PIXMAP(&tile[6], NULL, "road3.xpm");
    case (6):
     LOAD_GAME_PIXMAP(&tile[5], NULL, "road2.xpm");
	   
 	}

	chdir("../..");

	// Updates the SPRITE_SIZE
	if (rw != rh)
		dbg("%s's roads pixmap size is %d/%d (not a square!). Using the width as the sprite size", level.tileset, rw, rh);
	if (rw != SPRITE_SIZE || TILE_WIDTH != 10 || TILE_HEIGHT != 10) {
		SPRITE_SIZE = rw;
		resize_window(WINDOW_WIDTH, WINDOW_HEIGHT);
	}

	tile[FLAG] = create_pixmap(SPRITE_SIZE, SPRITE_SIZE);
	tile[PIT] = create_pixmap(SPRITE_SIZE, SPRITE_SIZE);

	blue_car.img[UP] = &blue_carImg[UP];
	blue_car.img[RIGHT] = &blue_carImg[RIGHT];
	blue_car.img[DOWN] = &blue_carImg[DOWN];
	blue_car.img[LEFT] = &blue_carImg[LEFT];

	// Create the flag tile
	blit(tile[ROAD], tile[FLAG], 0, 0, SPRITE_SIZE, SPRITE_SIZE);
	m_blit(flag.pix, flag.mask, tile[FLAG], 0, 0, SPRITE_SIZE, SPRITE_SIZE);

	return 1;
}

void draw_map(mapStruct *Map) {
	int value;
	int mx, my, mw, mh, sx, sy, sw, sh, b;
	int vL, vR, vT, vB;

	mw = Map->width;
	mh = Map->height;

	// Draw the map and the flags
	for (my = Map->tY-1; my < Map->tY + TILE_WIDTH + 1; my++) {
		for (mx = Map->tX-1; mx < Map->tX + TILE_HEIGHT + 1; mx++) {
			sx = toMap(mx-Map->tX) - Map->offX;
			sy = toMap(my-Map->tY) - Map->offY;
			sw = sh = SPRITE_SIZE;
			if (sx > MAP_WIDTH - SPRITE_SIZE)
				sw = MAP_WIDTH - sx;
			if (sy > MAP_HEIGHT - SPRITE_SIZE)
				sh = MAP_HEIGHT - sy;
			if (my >= 0 && mx >= 0 && my < mh && mx < mw) {
				// In the map, draw
				value = Map->data[my*mw + mx];
				blit(tile[value], dbuffer(), sx, sy, sw, sh);
				if (value == 1 && (b = level.border) > 0) {
					vL = mx > 0 ? Map->data[my*mw + mx - 1] : 1;
					vR = mx < (mw-1) ? Map->data[my*mw + mx + 1] : 1;
					vT = my > 0 ? Map->data[(my-1)*mw + mx] : 1;
					vB = my < (mh-1) ? Map->data[(my+1)*mw + mx] : 1;
					if (vL != 1)
						rect (dbuffer(), sx, sy, b, sh, black, 1);
					if (vR != 1)
						rect (dbuffer(), sx + sw - b, sy, b, sh, black, 1 );
					if (vT != 1)
						rect (dbuffer(), sx, sy, sw, b, black, 1);
					if (vB != 1)
						rect (dbuffer(), sx, sy + sh - b, sw, b, black, 1);
				}
			}
			else {
				// Out of the map
				blit(tile[OUT], dbuffer(), sx, sy, SPRITE_SIZE, SPRITE_SIZE);
			}
		}
	}
}

void draw_car(carStruct *car) {
	m_blit(car->img[car->direction]->pix, car->img[car->direction]->mask, dbuffer(), MIDDLE_X, MIDDLE_Y, SPRITE_SIZE, SPRITE_SIZE);
}

void process_input(KeySym key)
{
	switch(key)
	{
		case(XK_KP_4):
		case(XK_Left):
		case(XK_KP_Left):
			if (blue_car.ai>=0)
				changeDir(&blue_car, map, LEFT);
			break;
		case(XK_KP_6):
		case(XK_Right):
		case(XK_KP_Right):
			if (blue_car.ai>=0)
				changeDir(&blue_car, map, RIGHT);
			break;
		case(XK_KP_8):
		case(XK_Up):
		case(XK_KP_Up):
			if (blue_car.ai>=0)
				changeDir(&blue_car, map, UP);
			break;
		case(XK_KP_2):
		case(XK_Down):
		case(XK_KP_Down):
			if (blue_car.ai>=0)
				changeDir(&blue_car, map, DOWN);
			break;
		case(XK_Escape):
			ENDMAP = true;
			mapResult = RESULT_QUIT;
			ldbg("User has quit!");
			break;
		case (XK_KP_Space):
		case (XK_space):
			// He wants smoke!
			// Check if it can have smoke
			if (smoke.x[smoke.current] >= 0 && smoke.y[smoke.current] >= 0)
				break;
			// Ok, smoke it!
			smoke.x[smoke.current] = toMap(toTile(blue_car.x));
			smoke.y[smoke.current] = toMap(toTile(blue_car.y));
			actualFuel -= SMOKE_FUEL;
			smoke.left[smoke.current] = SMOKE_DURATION;
//					dbg("Smoke %d at: %d\t%d", smoke.current, smoke.x[0], smoke.y[0]);
			smoke.current++;
			if (smoke.current >= SMOKE_NUM)
				smoke.current = 0;
#ifdef USE_SOUND
			play_sound (SOUND_SMOKE);
#endif
			break;
		default: break;
	}
}

void draw_ai(aiStruct *ai) {
	int a;
	int sw, sh;
	int sx, sy;
	for (a = 0; a < ai->number; a++) {
		// Check if it is to draw the red car
		sx = ai->bot[a].x - (toMap(map->tX) + map->offX);
		sy = ai->bot[a].y - (toMap(map->tY) + map->offY);
		if (sy <= -SPRITE_SIZE || sx <= -SPRITE_SIZE || sy >= MAP_HEIGHT || sx >= MAP_WIDTH)
			continue;
		sw = (sx > MAP_WIDTH - SPRITE_SIZE) ? MAP_WIDTH - sx: SPRITE_SIZE;
		sh = (sy > MAP_HEIGHT - SPRITE_SIZE) ? MAP_HEIGHT - sy: SPRITE_SIZE;
		m_blit(ai->bot[a].img[ai->bot[a].direction]->pix, ai->bot[a].img[ai->bot[a].direction]->mask, dbuffer(), sx, sy, sw, sh);		
	}
}

void draw_smoke() {
	int a;
	int sx, sy, sw, sh;
	for (a = 0; a < SMOKE_NUM; a++) {
		// Check if it is to draw the smoke
		if (smoke.x[a] < 0 || smoke.y[a] < 0)
			continue;
		sx = smoke.x[a] + -(toMap(map->tX) + map->offX);
		sy = smoke.y[a] - (toMap(map->tY) + map->offY);
		if (sx <= -SPRITE_SIZE || sy <= -SPRITE_SIZE || sx >= MAP_WIDTH || sy >= MAP_HEIGHT)
			continue;
	  sw = (sx > MAP_WIDTH - SPRITE_SIZE) ? MAP_WIDTH - sx : SPRITE_SIZE;
		sh = (sy > MAP_HEIGHT - SPRITE_SIZE) ? MAP_HEIGHT - sy : SPRITE_SIZE;
		m_blit(smoke_tile.pix, smoke_tile.mask, dbuffer(), sx, sy, sw, sh);		
	}
	if (score.timer > 0 && score.timer != 128) {
		score.timer--;
		sx = score.x - (toMap(map->tX) + map->offX);
		sy = score.y - (toMap(map->tY) + map->offY) - text_height(FONT_SMALL);
		if (!(sx <= -SPRITE_SIZE || sy <= -SPRITE_SIZE || sx >= MAP_WIDTH || sy >= MAP_HEIGHT)) {
			sprintf(scrStr, "%d", score.last);
			text(dbuffer(), sx, sy, scrStr, white, FONT_SMALL);
			if (score.timer & 128)
				text(dbuffer(), sx, sy, "1 UP!", white, FONT_SMALL);
		}
	}
	if (crashed) {
			m_blit(crash_tile.pix, crash_tile.mask, dbuffer(), MIDDLE_X, MIDDLE_Y, SPRITE_SIZE, SPRITE_SIZE);		
	}
}

void clean_info_bar() {
	// Cause draw_map likes to put extra stuff
	rect(dbuffer(), MAP_WIDTH, 0, WINDOW_WIDTH - MAP_WIDTH+1, MAP_HEIGHT+1, black, 1);
}

/* The info bar should be this way:
 * (yeah, I suck in ASCII art)
 * |--------------|
 * | Lev name:    |
 * | Map: n / T   |
 * | Hi score:    |
 * |              |
 * | Lives:       |
 * | Score:       |
 * |              |
 * | Fuel:        |
 * | ----------   |
 * |              |
 * |   RADAR      |
 * |--------------| */
void draw_info_bar() {
  // Clean everything
	//rect(dbuffer(), SCORE_X, SCORE_Y, SCORE_WIDTH, SCORE_HEIGHT, black, 1);

  // Draw the level name
  text(dbuffer(), SCORE_X+(SCORE_WIDTH - text_width(level.name, FONT_SMALL))/2, 
       SCORE_Y + 2, level.name, green, FONT_SMALL);

  // Draw the map number / total
  text(dbuffer(), SCORE_X + 2, SCORE_Y + 2 + 20, mapString, green, FONT_SMALL);

  // Draw the hiscore
  if (hiString) {
    text(dbuffer(), SCORE_X + 2, SCORE_Y + 2 + 20 + 20, "Hiscore: ", green, FONT_SMALL);
    text(dbuffer(), SCORE_X - 4 + SCORE_WIDTH - text_width(hiString, FONT_SMALL),
          SCORE_Y + 2 + 20 + 20, hiString, green, FONT_SMALL);
  }

	// Draw the fuel
	text(dbuffer(), SCORE_X + 2, SCORE_Y + FUEL_Y - text_height(FONT_SMALL), "Fuel:", yellow, FONT_SMALL);
	if (actualFuel > 0)
		rect(dbuffer(), FUEL_X, FUEL_Y, (FUEL_WIDTH * actualFuel) / map->startFuel, FUEL_HEIGHT, yellow, 1);

#ifdef NUM_LIVES
	// Draw the lives
	text(dbuffer(), SCORE_X + 2, SCORE_Y + 2 + 20 + 20 + 20 + 20, livesString, white, FONT_SMALL);
#endif

	// Draw the score
	text(dbuffer(), SCORE_X + 2, SCORE_Y + 2 + 20 + 20 + 20 + 20 + 20, "Score:", white, FONT_SMALL);
	text(dbuffer(), SCORE_X - 4 + SCORE_WIDTH - text_width(score.string, FONT_SMALL), SCORE_Y + 2 + 20 + 20 + 20 + 20 + 20, score.string, white, FONT_SMALL);

}

bool in_radar(int x, int y, int carX, int carY, int range) {
	int cX = x > carX ? x - carX : carX - x; 
	int cY = y > carY ? y - carY : carY - y; 
	int g = range * range;
	return ((cX * cX + cY * cY) <= g);
}

void draw_radar(int mode, int range) {
	int a;
	int x, y;
  // Clean up
  //rect(dbuffer(), RADAR_X, RADAR_Y, RADAR_WIDTH, RADAR_HEIGHT, black, 1);
	// Draw background
	rect(dbuffer(), RADAR_X - 1, RADAR_Y - 1, radar.width + 2, radar.height + 2, green, 1);
	if (mode != ORIGINAL) {
		for (x = 0; x < map->width; x++) {
			for (y = 0; y < map->height; y++) {
				if (map->data[(y*map->width)+x] == BLOCK) {
					if (mode != SONAR || in_radar(toMap(x), toMap(y), blue_car.x, blue_car.y, range))
						rect(dbuffer(), RADAR_X + RADARBIT*x, RADAR_Y + RADARBIT*y, RADARBIT, RADARBIT, black, 1);
				}
			}
		}
	}
	rect(dbuffer(), RADAR_X + (int)(RADARBIT*((float)blue_car.x/SPRITE_SIZE)), RADAR_Y + (int)(RADARBIT*((float)blue_car.y/SPRITE_SIZE)), RADARBIT, RADARBIT, blue, 1);
	for (a = 0; a < map->flagNum; a++) {
		if (map->flagX[a] >= 0 && map->flagY[a] >= 0)
			if (mode != SONAR || in_radar(toMap(map->flagX[a]), toMap(map->flagY[a]), blue_car.x, blue_car.y, range))
				rect(dbuffer(), RADAR_X + RADARBIT*map->flagX[a], RADAR_Y + map->flagY[a]*RADARBIT, RADARBIT, RADARBIT, yellow, 1);
	}
	for (a = 0; a < AI.number; a++) {
		if (mode != SONAR || in_radar(AI.bot[a].x, AI.bot[a].y, blue_car.x, blue_car.y, range))
			rect(dbuffer(), RADAR_X + (int)(RADARBIT*((float)AI.bot[a].x/SPRITE_SIZE)), RADAR_Y + (int)(RADARBIT*((float)AI.bot[a].y/SPRITE_SIZE)), RADARBIT, RADARBIT, red, 1);
	}
}

void draw_light(int light)
{

	circle(dbuffer(), (MAP_WIDTH-100), 30, 32, 32, white, 1);
	circle(dbuffer(), (MAP_WIDTH-100)+34, 30, 32, 32, white, 1);
	circle(dbuffer(), (MAP_WIDTH-100)+68, 30, 32, 32, white, 1);

	// Draw first light (red)
	circle(dbuffer(), (MAP_WIDTH-100)+1, 31, 30, 30, red, 1);
	
	// Draw second light (yellow)
	circle(dbuffer(), (MAP_WIDTH-100)+35, 31, 30, 30, (light>=1)?yellow:black, 1);

	// Draw third light (green)
	circle(dbuffer(), (MAP_WIDTH-100)+69, 31, 30, 30, (light>=2)?green:black, 1);

	// Draw "bonus" title if necessary
	if (map->bonus)
		text(dbuffer(), (MAP_WIDTH-text_width("BONUS!", FONT_BIG))/2, (MAP_HEIGHT-text_height(FONT_BIG)-100)/2, "BONUS!", yellow, FONT_BIG);

}

void fix_geometry(int size) {
	/*
	SPRITE_SIZE = size;
	//handleGeometry();
	MAP_WIDTH+=134;
	MAP_HEIGHT+=4;
	if (buffersCreated) {
		XFreePixmap(dpy, radar.pix);
		XFreePixmap(dpy, scoreBuffer);		
		XFreePixmap(dpy, dbuffer);
		buffersCreated = false;
	}
	ldbg("Geometry fixed: MAP_WIDTH: %d \tMAP_HEIGHT: %d\nWIN_WIDTH: %d\tWIN_HEIGHT %d", MAP_WIDTH, MAP_HEIGHT, WIN_WIDTH, WIN_HEIGHT);
	initBuffers();
	*/
}

void init_screen() {

	blue = get_color(COLOR_BLUE);
	red = get_color(COLOR_RED);
	green = get_color(COLOR_GREEN);
	black = get_color(COLOR_BLACK);
	yellow = get_color(COLOR_YELLOW);
	white = get_color(COLOR_WHITE);

}

void draw_window(int x, int y, int w, int h) {
	unsigned long color = get_color("rgb:BB/BB/BB");
	XImage *blend_window;
	// Draw the borders (5 is the border size)
	rect(dbuffer(), x, y, w, 5, color, 1);
	rect(dbuffer(), x, y, 5, h, color, 1);
	rect(dbuffer(), x, y+h-5, w, 5, color, 1);
	rect(dbuffer(), x+w-5, y, 5, h, color, 1);
	// Draw the background
	blend_window = blend_image(dbuffer(), x+5, y+5, w-10, h-10, &blend_cache);
	i_blit(blend_window, dbuffer(), x+5, y+5, w-10, h-10);
	//XSetForeground(dpy, windowGC, get_color(COLOR_BACK));
	//XFillRectangle(dpy, dbuffer, windowGC, x+5, y+5, w-10, h-10);
}
