/***************************************************************************
                          graphics.c - XRally Graphics API
                             -------------------
    begin                : Sat Jan 1 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: graphics.c,v 1.15 2000/12/07 12:44:10 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "graphics.h"
#include "global.h"
#include <sys/time.h>

// Internal functions

static char ** _split_text(const char *_txt);
static int _check_event(XEvent *xev);
long time_diff(struct timeval *rt);

// Basic X11 variables
static Window win;
static Colormap cmap;
static Display *dpy;
static GC textGC;
static GC tileGC; 
static GC solidGC;
static Pixmap double_buffer; // Double Buffer
static int depth;
static int screen;
static XFontStruct *ft[4];

static unsigned long valuemask = 0;
static XGCValues values;

static XSizeHints size_hints;

// Internal window width/height
static int _WINDOW_WIDTH = DEFAULT_WINDOW_WIDTH;
static int _WINDOW_HEIGHT = DEFAULT_WINDOW_HEIGHT;

#define WINDOW_NAME PACKAGE " " VERSION

int init_display (char *display) {
 	char *window_name = WINDOW_NAME;
	char *icon_name = ICON_NAME;
	XSetWindowAttributes att;
	Window rootwin;

	dpy = XOpenDisplay(NULL);

	if (dpy == NULL)
	{
		hdbg("Failed to open display");
		return false;
	}

	screen = DefaultScreen(dpy);
	depth = DefaultDepth(dpy, screen);
	cmap = DefaultColormap(dpy, screen);

	ft[FONT_SMALL] = XLoadQueryFont(dpy, "-*-helvetica-*-r-*-*-14-*-*-*-*-*-*-*");
	ft[FONT_MEDIUM] = XLoadQueryFont(dpy, "-*-helvetica-*-r-*-*-18-*-*-*-*-*-*-*");
	ft[FONT_BIG] = XLoadQueryFont(dpy, "-*-helvetica-*-r-*-*-24-*-*-*-*-*-*-*");
	ft[FONT_FIXED] = XLoadQueryFont(dpy, "-bitstream-courier-bold-r-*-*-18-*-*-*-*-*-*-*");
	if (ft[FONT_FIXED] == NULL)
		ft[FONT_FIXED] = XLoadQueryFont(dpy, "-*-courier-bold-r-*-*-18-*-*-*-*-*-*-*");


	if (ft[FONT_SMALL] == NULL || ft[FONT_MEDIUM] == NULL || ft[FONT_BIG] == NULL || ft[FONT_FIXED] == NULL) {
		hdbg("You don't have the helvetica or the fixed font!!!!! That's too bad!!! Improvising...");
		ft[FONT_SMALL] = XLoadQueryFont(dpy, "-*-*-*-*-*-*-14-*-*-*-*-*-*-*");
		ft[FONT_MEDIUM] = XLoadQueryFont(dpy, "-*-*-*-*-*-*-18-*-*-*-*-*-*-*");
		ft[FONT_BIG] = XLoadQueryFont(dpy, "-*-*-*-*-*-*-24-*-*-*-*-*-*-*");
		ft[FONT_FIXED] = XLoadQueryFont(dpy, "-*-*-*-*-*-*-24-*-*-*-*-*-*-*");
	}	

	rootwin = RootWindow(dpy, screen);
	win = XCreateSimpleWindow(dpy, rootwin, 10, 10, DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT, 5,	BlackPixel(dpy, screen), get_color(COLOR_BACK));
	att.event_mask = KeyPressMask | ExposureMask;
	XChangeWindowAttributes(dpy, win, CWEventMask, &att);

	XSetWindowColormap(dpy, win, cmap);


	size_hints.flags = PSize | PMinSize | PMaxSize;
	size_hints.min_width = DEFAULT_WINDOW_WIDTH;
	size_hints.max_width = DEFAULT_WINDOW_WIDTH;
	size_hints.min_height = DEFAULT_WINDOW_HEIGHT;
	size_hints.max_height = DEFAULT_WINDOW_HEIGHT;

	XSetStandardProperties(dpy, win, window_name, icon_name, None, 0, 0, &size_hints);

	XSelectInput(dpy, win, KeyPressMask | ExposureMask);
	XMapWindow(dpy, win);

	double_buffer = XCreatePixmap(dpy, win, DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT, depth);
	tileGC = XCreateGC(dpy, double_buffer, valuemask, &values);
	solidGC = XCreateGC(dpy, double_buffer, valuemask, &values);
	textGC = XCreateGC(dpy, win, valuemask, &values);
	XSetGraphicsExposures(dpy, tileGC, false);
	XSetGraphicsExposures(dpy, solidGC, false);
	XSetFillStyle(dpy, tileGC, FillTiled);
	XSetFillStyle(dpy, solidGC, FillSolid);

	return true;
}

int load_pixmap(Pixmap *pix, Pixmap *mask, char *filename, int *width, int *height){
	
	int a;
	XpmAttributes attributes;

	attributes.valuemask = XpmColormap | XpmCloseness;
	attributes.colormap = cmap;
	attributes.closeness = 65535;

	a =	XpmReadFileToPixmap(dpy, win, filename, pix, mask, &attributes);

	if (width)
		*width = attributes.width;
	if (height)
		*height = attributes.height;

	return a;
}

void blit(Pixmap src, Drawable dest, int x, int y, int w, int h) {

	//XSetTile(dpy, tileGC, src);
	//XSetTSOrigin(dpy, tileGC, x, y);
	XSetClipMask(dpy, tileGC, None);
	XCopyArea(dpy, src, dest, tileGC, 0, 0, w, h, x, y);
	//XFillRectangle(dpy, dest, tileGC, x, y, w, h);

}

void us_blit(Pixmap src, Drawable dest, int x, int y, int w, int h)
{
#define SEGSIZE 16
#define LINESIZE 2
	int i, j;
	int seg = h / SEGSIZE;		/* Number of segments */

	XSetTile(dpy, tileGC, src);
	XSetTSOrigin(dpy, tileGC, x, y);
	XSetClipMask(dpy, tileGC, None);

	for (j = seg + (SEGSIZE / LINESIZE) - 1; j >= 0; j--) {
		for (i = 0; i < seg; i++) {
			int y1 = y + j * LINESIZE + i * SEGSIZE;
			int h1 = LINESIZE;

			if (y1 > (y + h) || i > j || i < (j - (SEGSIZE / LINESIZE)))
				continue;

			XFillRectangle(dpy, dest, tileGC, x, y1, w, h1);
		}
		if (check_key() != XK_VoidSymbol)
			break;
		graphics_sync ();
	}
	blit(src, dest, x, y, w, h);
}

void usi_blit(XImage *src, Drawable dest, int x, int y, int w, int h)
{
	int i, j;
	int seg = h / SEGSIZE;		/* Number of segments */

	XSetClipMask(dpy, tileGC, None);

	for (j = seg + (SEGSIZE / LINESIZE) - 1; j >= 0; j--) {
		for (i = 0; i < seg; i++) {
			int dy = j * LINESIZE + i * SEGSIZE;
			int y1 = y + j * LINESIZE + i * SEGSIZE;
			int h1 = LINESIZE;

			if (y1 > (y + h) || i > j || i < (j - (SEGSIZE / LINESIZE)))
				continue;

			XPutImage(dpy, dest, tileGC, src, 0, dy, x, y1, w, h1);
		}
		if (check_key() != XK_VoidSymbol)
			break;
		graphics_sync ();
	}
	i_blit(src, dest, x, y, w, h);
}

void s_blit(Pixmap src, Drawable dest, int x, int y, int w, int h)
{
	int i, j;
	int seg = h / SEGSIZE;		/* Number of segments */

	XSetTile(dpy, tileGC, src);
	XSetTSOrigin(dpy, tileGC, x, y);
	XSetClipMask(dpy, tileGC, None);

	for (j = 0; j < seg + (SEGSIZE / LINESIZE); j++) {
		for (i = 0; i < seg; i++) {
			int y1 = y + j * LINESIZE + i * SEGSIZE;
			int h1 = LINESIZE;

			if (y1 > (y + h) || i > j || i < (j - (SEGSIZE / LINESIZE)))
				continue;

			XFillRectangle(dpy, dest, tileGC, x, y1, w, h1);
		}
		if (check_key() != XK_VoidSymbol)
			break;
		graphics_sync ();
	}
	blit(src, dest, x, y, w, h);
}

void m_blit(Pixmap srcPix, Pixmap srcMask, Drawable dest, int x, int y, 
		        int w, int h) {

	XSetTile(dpy, tileGC, srcPix);
	XSetTSOrigin(dpy, tileGC, x, y);
	XSetClipOrigin(dpy, tileGC, x, y);
	XSetClipMask(dpy, tileGC, srcMask);
	XFillRectangle(dpy, dest, tileGC, x, y, w, h);

}
	
void i_blit(XImage *src, Drawable dest, int x, int y, int w, int h) {

	XSetClipMask(dpy, tileGC, None);
	XPutImage(dpy, dest, tileGC, src, 0, 0, x, y, w, h);

}

void si_blit(XImage *src, Drawable dest, int x, int y, int w, int h) {

	int i, j;
	int seg = h / SEGSIZE;		/* Number of segments */

	XSetClipMask(dpy, tileGC, None);

	for (j = 0; j < seg + (SEGSIZE / LINESIZE); j++) {
		for (i = 0; i < seg; i++) {
			int dy = j * LINESIZE + i * SEGSIZE;
			int y1 = y + dy;
			int h1 = LINESIZE;

			if (y1 > (y + h) || i > j || i < (j - (SEGSIZE / LINESIZE)))
				continue;

			XPutImage(dpy, dest, tileGC, src, 0, dy, x, y1, w, h1);
		}
		if (check_key() != XK_VoidSymbol)
			break;
		graphics_sync ();
	}
	i_blit(src, dest, x, y, w, h);
}

void rect(Drawable dest, int x, int y, int w, int h, unsigned long color_index,
					int filled) {
	XSetForeground(dpy, solidGC, color_index);
	if (filled)
		XFillRectangle(dpy, dest, solidGC, x, y, w, h);
	else
		XDrawRectangle(dpy, dest, solidGC, x, y, w, h);

}

void point(Drawable dest, int x, int y, unsigned long color_index) {
	XSetForeground(dpy, solidGC, color_index);
	XDrawPoint(dpy, dest, solidGC, x, y);

}

int text_width(char *txt, int font) { 
	return XTextWidth(ft[font], txt, strlen(txt)); 
};

int text_height(int font) {
 	return (ft[font]->max_bounds.ascent + ft[font]->max_bounds.descent);
};

static char ** _split_text(const char *_txt) {
	char **itens;
	char *txt;
	int count = 1;
	int a;
	txt = (char *)malloc( (strlen(_txt) + 2) * sizeof(char) );
	strcpy(txt, _txt);
	ldbg("pre-split: %s", txt);
	for (a = 0; a < strlen(txt); a++) {
		if (txt[a] == '\n')
			count++;
	}
	ldbg("lines count: %d", count);
	itens = (char **)calloc(count + 1, sizeof(char *));
	for (a = 0; a < (count+1); a++) {
		itens[a] = txt;
		while (*txt != '\n' && *txt != 0) {
			txt++;
		}
		if (*txt == 0) {
			break;
		}
		*txt = 0;
		txt++;
	}
	itens[count] = 0;
	return itens;
}

void text(Drawable dest, int x, int y, const char *text_string, 
		 unsigned long color_index, int font) {

	int a;
	char **text = _split_text(text_string);
	XSetForeground(dpy, textGC, color_index);
	XSetFont(dpy, textGC, ft[font]->fid);
	for (a = 0; text[a] != 0; a++) {
		ldbg("post-split: %s", text[a]);
		XDrawString(dpy, dest, textGC, x, y+a*text_height(font)+ft[font]->ascent, text[a], strlen(text[a]));
	}
	free(text[0]);
	free(text);

}

void circle(Drawable dest, int x, int y, int w, int h, 
						unsigned long color_index, int filled) {
	XSetForeground(dpy, solidGC, color_index);
	if (filled)
		XFillArc(dpy, dest, solidGC, x, y, w, h, 0, 23040);
	else
		XDrawArc(dpy, dest, solidGC, x, y, w, h, 0, 23040);

}

XImage *get_image(Pixmap src, int x, int y, int w, int h) {
	return XGetImage(dpy, src, x, y, w, h, AllPlanes, ZPixmap);
}

unsigned long get_color(const char *color_string) {
	XColor color;
	XAllocNamedColor(dpy, cmap, color_string, &color, &color);
	return color.pixel;
}

void alloc_color(XColor *color) {
	XAllocColor(dpy, cmap, color);
}

void query_color(XColor *color) {
	XQueryColor(dpy, cmap, color);
}

void resize_window(int width, int height) {
	dbg("Resizing window to %d / %d", width, height);
	size_hints.flags = PSize | PMinSize | PMaxSize;
	size_hints.min_width = width;
	size_hints.max_width = width;
	size_hints.min_height = height;
	size_hints.max_height = height;
	XSetStandardProperties(dpy, win, WINDOW_NAME, ICON_NAME, None, 0, 0l, &size_hints);
	XResizeWindow(dpy, win, width, height);
	_WINDOW_WIDTH = width;
	_WINDOW_HEIGHT = height;
	XFreePixmap(dpy, double_buffer);
	double_buffer = XCreatePixmap(dpy, win, _WINDOW_WIDTH, _WINDOW_HEIGHT, depth);
	return;
}

Pixmap create_pixmap(int width, int height) {
	return XCreatePixmap(dpy, double_buffer, width, height, depth);
}

void graphics_sync() {
	//blit(double_buffer, win, 0, 0, 480, 480);
	//blit(double_buffer, win, 484, 242, 128, 240);
	//blit(double_buffer, win, 482, 174, 128, 10);    

	XSetClipMask(dpy, tileGC, None);
	XCopyArea(dpy, double_buffer, win, tileGC, 0, 0, _WINDOW_WIDTH, _WINDOW_HEIGHT, 0, 0);
	//XCopyArea(dpy, double_buffer, win, tileGC, 0, 0, 480, 480, 0, 0);
	XSync(dpy, false);
}

void close_display() {
	XCloseDisplay(dpy);
}

Pixmap dbuffer() {
	return double_buffer;
}

static int _check_event(XEvent *xev) {
	XEvent *last = NULL;

	while (XCheckMaskEvent(dpy, KeyPressMask | ExposureMask, xev)) {
		switch (xev->type) {
			case KeyPress:
				last = xev;
				break;
			case Expose:
				XSetClipMask(dpy, tileGC, None);
				XCopyArea(dpy, double_buffer, win, tileGC, xev->xexpose.x, xev->xexpose.y, xev->xexpose.width, xev->xexpose.height, xev->xexpose.x, xev->xexpose.y);
				XSync(dpy, false);
				break;
			default:
				break;
		}
	}
	// No events
	xev = last;
	return (last) ? true : false;
	//return false;
}

KeySym check_key() {
	XEvent xev;
	if (_check_event(&xev) && xev.type == KeyPress)
		return XLookupKeysym(&xev.xkey, 0);
	// No events
	return XK_VoidSymbol;
}

KeySym wait_key() {
	KeySym key = XK_VoidSymbol;
	while (key == XK_VoidSymbol) {
		key = check_key();
	}
	return key;
}

void wait_time(long time_wait) {
	struct timeval it;
	XEvent xev;
	gettimeofday(&it, NULL);
	while (time_diff(&it) < time_wait)
		_check_event(&xev);
}

KeySym wait_char(char *str) {
	KeySym key = XK_VoidSymbol;
	XEvent xev;
	while (!_check_event(&xev) || xev.type != KeyPress);
	XLookupString(&xev.xkey, str, 3, &key,  NULL);
	return key;
}

void text_rectangle(const char *txt, int font, int *w, int *h) {

	char **text = _split_text(txt);
	int a = 0;
	*w = text_width(text[a], font);
	for (a = 0; text[a] != 0; a++)
		*w = text_width(text[a], font) > *w ? text_width(text[a], font) : *w;
	*h = a * text_height(font);

}
