/***************************************************************************
                          level.c  -  description
                             -------------------
    begin                : Sat Jan 1 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: level.c,v 1.8 2001/01/01 18:49:16 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "config.h"
#include "level.h"
#include "menu.h"
#include "map.h"
#include "graphics.h"
#include "game.h"

#ifdef USE_SOUND
#include "sound.h"
#endif

bool load_sound = true;
bool loaded_sound = false;		/* Fixes bug #122897 */

extern carStruct blue_car;

extern int TILE_WIDTH;
extern int TILE_HEIGHT;

bool quickLoadLevel(levelStruct *level, char *filename) {
	char buf[256];
	long pos = 0;
	unsigned long int sum = 0;
		
	// Check if the file is readable, exists, etc

	if( chdir("maps") < 0 ) {
		hdbg("Can't find subdirectory 'maps'");
		return false;
	}

	if( access(filename, R_OK) < 0 ) {
		perror("Access");
		hdbg("Error trying to access \"%s\"", filename);
		return false;
	}

	if( !(level->file = fopen(filename, "r")) ) {
		hdbg("Error loading map \"%s\"!", filename);
		fclose(level->file);
		return false;
	}

	// Return to current directory, so that isn't any problem with sound
	chdir("..");

	// Init the level
	initLevel(level);

	// Discover the number of maps, so that it can allocate memory for them
	while (fgets(buf, 256, level->file)) {
		// Comment
		if (*buf == '#')
			continue;
		// Wasn't a comment, so add it to the sum
		sum += checksum(buf);
		// Map!
		if (*buf == '-') {
			// Map!
			// Set up a file descriptor
			// Increase the map counter
      if (level->number == 0)
			  pos = ftell(level->file);
			level->number++;
		}
	}
	rewind(level->file);

	ldbg("*** LEVEL LOADING STARTED ***\nlevel has %d maps\n---", level->number);

  // Load memory for the first map
  ldbg("Allocing memory for the first map");
  level->map = (mapStruct *)calloc(1, (size_t)sizeof(mapStruct));
  initMap(&level->map[0]);
	
	// Start scanning through the global options
	*buf = 0;
	// Don't want to load the sound
	load_sound = false;
	while (*buf != '-') {
		if (!fgets(buf, 256, level->file)) {
			// level reached end before a map starts!!!
			ldbg("level doesn't have any map!");
			return false;
		}
		// Ignore comments
		if (*buf == '#')
			continue;
		parseGlobalOption(getOptionNumber(buf), buf, level, 1);	
	}

  // Parse the first map
  parseMapOptions(level, pos, &level->map[0], 0);

	// Can load the sound again
	load_sound = true;

	// Save checksum
	sprintf(level->sum, "%lu", sum);
	ldbg("Map checksum: %lu", sum);

	ldbg("*** LEVEL LOADING ENDED ***\n---");

	return true;
}


bool loadLevel(levelStruct *level, char *filename) {
	// This function parses the level file and call the 
	// necessary functions.
	// 1) Gets the number of maps, malloc them and add a fd for each
	// 2) Run the global options (including map options in global mode)
	// 3) Run the map options

	char buf[256];
  struct node_s {
    long data;
    struct node_s *prev;
  } *node = NULL, *node_tmp = NULL;
  long *pos;
	int a;
	unsigned long int sum = 0;
		
	// Check if the file is readable, exists, etc

	if( chdir("maps") < 0 ) {
		hdbg("Can't find subdirectory 'maps'");
		return false;
	}

	update_splash();

	if( access(filename, R_OK) < 0 ) {
		perror("Access");
		hdbg("Error trying to access \"%s\"", filename);
		return false;
	}

	update_splash();

	if( !(level->file = fopen(filename, "r")) ) {
		hdbg("Error loading map \"%s\"!", filename);
		fclose(level->file);
		return false;
	}

	update_splash();

	// Return to current directory, so that isn't any problem with sound
	chdir("..");

	// Init the level
	initLevel(level);

	update_splash();

	// Discover the number of maps, so that it can allocate memory for them
	while (fgets(buf, 256, level->file)) {
		// Comment
		if (*buf == '#')
			continue;
		// Wasn't a comment, so add it to the sum
		sum += checksum(buf);
		// Map!
		if (*buf == '-') {
			// Map!
			// Set up a file descriptor
			// Increase the map counter
      node_tmp = (struct node_s *)malloc(sizeof(struct node_s));
      node_tmp->prev = node;
      node_tmp->data = ftell(level->file);
      node = node_tmp;
			level->number++;
		}
	}
	rewind(level->file);

  pos = (long*)malloc(level->number*sizeof(long));
  node_tmp = node;
  for (a = level->number-1; a >= 0; a--) {
    pos[a] = node->data;
    node = node->prev;
  }

	// Alloc memory for the maps
	level->map = (mapStruct *)calloc(level->number, (size_t)sizeof(mapStruct));
	// Init map variables
	for (a = 0; a < level->number; a++) {
		initMap(&level->map[a]);
	}

	update_splash();

	ldbg("*** LEVEL LOADING STARTED ***\nlevel has %d maps\n---", level->number);
	
	// Start scanning through the global options
	*buf = 0;
	loaded_sound = false;
	while (*buf != '-') {
		if (!fgets(buf, 256, level->file)) {
			// level reached end before a map starts!!!
			ldbg("level doesn't have any map!");
			return false;
		}
		// Ignore comments
		if (*buf == '#')
			continue;
		parseGlobalOption(getOptionNumber(buf), buf, level, 0);	
	}

	update_splash();

	if (!loaded_sound) {
		dbg ("Loading default sound file!");
		sound_load ("default.mod");
		level->volume = 0;
	}

	// Now parse through all the map options
	for (a = 0; a < level->number; a++) {
		ldbg("*--- Starting map %d --*", a);
		parseMapOptions(level, pos[a], &level->map[a], 1);
		ldbg("*--- Finished map %d --*", a);
	}

	update_splash();
	
	// End
  /*
	free(pos);
  node = node_tmp;
  while (node) {
    free(node);
    node = node->prev;
  }
  */

	// Save checksum
	sprintf(level->sum, "%lu", sum);
	ldbg("Map checksum: %lu", sum);

	update_splash();

	ldbg("*** LEVEL LOADING ENDED ***\n---");

	return true;
}

unsigned long int checksum(char *buf) {
	unsigned long int a=0;
	unsigned int b = 0x17;
	unsigned char *c;
	unsigned long int d=0;
	unsigned int pos = 1;
	
	c = buf;
	while (*c != 0) {
		d = *c << 8;
		a+=(d%(pos+b));
		b = *c;
		c++;
		pos++;
	}
	
	return a;	

}

int getOptionNumber(char *opt) {
	int a = -1;
	char cmd[256];
	sscanf(opt, "%s", cmd);
	// Check each one of the known cmds
	if (match("name"))
		a = NAME;
	if (match("desc"))
		a = DESC;
	if (match("tileset"))
		a = TILESET;
	if (match("size"))
		a = SIZE;
	if (match("speed"))
		a = SPEED;
	if (match("lives"))
		a = LIVES;
	if (match("fuel"))
		a = FUEL;
	if (match("bonus"))
		a = BONUS;
	if (match("fuelpts"))
		a = FUELPTS;
	if (match("radarmode"))
		a = RADARMODE;
	if (match("carai"))
		a = CARAI;
	if (match("border"))
		a = BORDER;
	if (match("sound"))
		a = SOUND;
	if (match("enemywait"))
		a = ENEMYWAIT;
  if (match("resol"))
    a = RESOL;
	return a;
}

void parseGlobalOption(int option, char *arg, levelStruct *level, int quickParseMap) {
	int a = 0;
	if (option < 0)
		return;
	if (arg[strlen(arg)-1] == '\n')
		arg[strlen(arg)-1] = 0;
	ldbg("Parsing global option %x: %s", option, arg);
	// Only the args... not command name please (and none of these naughty line breaks)
	arg = cut(arg);

	if (option & 0x80) {
		// Map option!
		for (a = 0; a < level->number; a++) {
      if (!quickParseMap || a == 0)
			  mapOption(option, arg, &level->map[a]);
		}
	} else if (option >= 0) {
		// Global option!
		levelOption(option, arg, level);
	} else {
		ldbg("Hey, what is a negative option?");
	}
	ldbg("---");
}

void parseMapOptions(levelStruct *level, long position, mapStruct *map, int full_loading) {
	char buf[256] = "";
	bool first = true;
	int a;
	// Go to the map position
	fseek(level->file, position, SEEK_SET);
	// Get each line
	while (fgets(buf, 256, level->file)) {
		// Ignore comments and the map end
		if (*buf == '#') 
			continue;
		if (*buf == '-')
			break;
		// Ok. Check if it's a command
	  a = getOptionNumber(buf);
		if (a < 0 || !(a & 0x80)) {
			// Not a command. Check if it's of the right width (the -1 is because of the LF/CR)
			if (strlen(buf)-1 == map->width) {
				// That's it!! That's map data
		    // Check if 'data' was malloced
				if (!map->data) {
					ldbg("Mallocing memory for the map data: %d bytes", map->height * map->width);
					map->data = (char *)malloc(map->width * map->height * sizeof(char) + 1 );
					strcpy(map->data, "");
					first = false;
				}
				// Copy data
				strncat(map->data, buf, map->width);
				continue;
			} else {
			ldbg("Not a map option: %d - %s(Len: %d, Width: %d)", a, buf, strlen(buf), map->width);
			continue;
			}
		}
		// Ok, it's a know option! Cut it and send!
		ldbg("Parsing map option %d: %s", a, buf);
		strncpy(buf, cut(buf), 256);
		mapOption(a, buf, map);
		ldbg("---");
	}

	// All map options were parsed and the data was filled.
	// Interpret the data now
  if (full_loading)
	  interpretData(map, level);
	
}

void mapOption(int option, char *arg, mapStruct *map) {
	int a = -1, b = -1;
	ldbg("Map argument: %s", arg); 
	switch(option) {
		case SIZE:
			sscanf(arg, "%d %d", &a, &b);
			map->width = a;
			map->height = b;
			break;
		case SPEED:
			sscanf(arg, "%d %d", &a, &b);
			if (a >= 0)
				map->carVel = a;
			if (b >= 0)
				map->aiVel = b;
			break;
		case FUEL:
			sscanf(arg, "%d %d", &a, &b);
			if (a >= 0)
				map->startFuel = a;
			if (b >= 0)
				map->fuelImpact = b;
			break;
		case BONUS:
			map->bonus = true;
			break;
		case FUELPTS:
			sscanf(arg, "%d", &a);
			map->fuelpts = a;
			break;
		case RADARMODE:
			sscanf(arg, "%d %d", &a, &b);
			map->radarmode = a;
			if (b >= 0)
				map->radarrange = b;	
			break;
		case ENEMYWAIT:
			sscanf(arg, "%d", &a);
			if (a >= 0)
				map->enemywait = a;
			break;
	}
}

void levelOption(int option, char *arg, levelStruct *level) {
	int a = -1, b = -1, c = -1;
	char d[80];
	ldbg("level argument: %s", arg);
	switch(option) {
		case NAME:
			// Copy the arg as the name
			strncpy(level->name, arg, 256);
			break;
		case DESC:
			// Don't do anything yet
			break;
		case TILESET:
			// Copy the tileset name
			strncpy(level->tileset, arg, 256);
			break;
		case LIVES:
			// Set the number of lives
			switch (sscanf(arg, "%d %d %d", &a, &b, &c)) {
				case 3:
					level->livesbonus = c;
				case 2:
					level->extra = b;
				case 1:
					level->lives = a;
			}
			if (level->lives <= 0)
				level->extra = 0;
			ldbg("Lives: %d\tExtra: %d\tLives bonus: %d", level->lives, level->extra, level->livesbonus);
			break;
		case CARAI:
			if (sscanf(arg, "%d", &a) >= 1)
				blue_car.ai = a;
			break;
		case BORDER:
			sscanf(arg, "%d", &a);
			level->border = a;
			break;
		case SOUND:
			if (load_sound) {
				sscanf(arg, "%s %d", d, &a);
				dbg ("load sound: %s... ", d);
			#ifdef USE_SOUND
				if (sound_load (d) < 0) {
					dbg ("error!");
					break;
				}
			#endif
				dbg ("ok");
				level->volume = a;
				loaded_sound = true;
			}
			break;
    case RESOL:
      sscanf(arg, "%d %d", &a, &b);
      if (a > 0)
        TILE_WIDTH = a;
      if (b > 0)
        TILE_HEIGHT = b;
      break;
		default:
			dbg("Option not supported");
			break;
	}
}

void initLevel(levelStruct *level) {
	level->number = 0;
	level->pics = PIXS;
	level->lives = NUM_LIVES;
	level->extra = EXTRA_LIVE;
	level->livesbonus = -1 * EXTRA_LIVE / NUM_LIVES;
	level->border = 0;
#ifdef USE_SOUND
	level->volume = 0;
#endif
  level->map = NULL;
  TILE_WIDTH = 10;
  TILE_HEIGHT = 10;
	blue_car.ai = 0;
}

void interpretData(mapStruct *map, levelStruct *level) {
	char *c;
	int x, y;
	int redNum = 0, flagNum = 0;
	// Debug information
	ldbg("Map data:\n%s", map->data);
	// Discover the redNum and FlagNum
	c = map->data;
	for (y = 0; y < (map->height * map->width); y++) {
		if (*c == 'A' || *c == 'a')
			redNum++;
		if (*c == 'F' || *c == 'f')
			flagNum++;
		c++;
	}
	// Alloc memory for the AI start positions and the flag positions
	if (redNum > 0) {
		map->red_startX = (int *)calloc(redNum, sizeof(int));
		map->red_startY = (int *)calloc(redNum, sizeof(int));
	}
	if (flagNum > 0) {
		map->flagX = (int *)calloc(flagNum, sizeof(int));
		map->flagY = (int *)calloc(flagNum, sizeof(int));
	}
	// Clean up the flagNum and redNum
	redNum = 0;
	flagNum = 0;
	// Interpret each data
	c = map->data;
	for (y = 0; y < map->height; y++) {
		for (x = 0; x < map->width; x++) {
			switch (*c) {
				case '0':
				case '1':
					*c-=48;
					break;
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':  // '2' = 50 ; 50 - 45 = 5 (custom passable)
					*c-=45;
					if (level->pics < (*c+1))
						level->pics = (*c+1);
					break;
				case 'z':
				case 'Z':
					*c=10;
					if (level->pics < (*c+1))
						level->pics = (*c+1);
					break;
				case 'x':
				case 'X':
					*c=11;
          if (level->pics < (*c+1))
          	level->pics = (*c+1);
					break;
				case 'c':			
				case 'C':
					*c=12;
          if (level->pics < (*c+1))
          	level->pics = (*c+1);
          break;
 				case 'v':
				case 'V':
					*c=13;
          if (level->pics < (*c+1))
          	level->pics = (*c+1);
          break;
 				case 'b':
				case 'B':
					*c=14;
          if (level->pics < (*c+1))
          	level->pics = (*c+1);
          break;
 				case 'd':
				case 'D':
				case 's':
				case 'S':
					*c = 0;
					map->startX = x;
					map->startY = y;
					map->startDir = DOWN;
					break;
				case 'u':
				case 'U':
					*c = 0;
					map->startX = x;
					map->startY = y;
					map->startDir = UP;
					break;
				case 'r':
				case 'R':
					*c = 0;
					map->startX = x;
					map->startY = y;
					map->startDir = RIGHT;
					break;
				case 'l':
				case 'L':
					*c = 0;
					map->startX = x;
					map->startY = y;
					map->startDir = LEFT;
					break;
				case 'a':
				case 'A':
					*c = 0;
					map->red_startX[redNum] = x;
					map->red_startY[redNum] = y;
					redNum++;
					break;
				case 'f':
				case 'F':
					*c = FLAG;
					map->flagX[flagNum] = x;
					map->flagY[flagNum] = y;
					flagNum++;
					break;
				case 'p':
				case 'P':
					*c = PIT;
					break;
				default:
					ldbg("Symbol unknow: %c\tAssumed 0", *c);
					*c = 0;
					break;
			}
			c++;
		}
  }
	map->redNum = redNum;
	map->flagNum = flagNum;
	map->flagLeft = flagNum;
	if (map->startX == -1 || map->startY == -1 || map->startDir == -1 ) {
		ldbg("Map doesn't have a starting position. Assuming 0,0");
		map->startX = 0;
		map->startY = 0;
		map->startDir = DOWN;
	} 
}

char *cut(char *arg) {
	int a = 0;
	while (*arg != ' ') {
		arg++;
	}
	arg++;
	do {
		if (arg[a] == 13 || arg[a] == 10)
			arg[a] = 0;
		a++;
	} while (arg[a] != 0);
	return arg;
}
