/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Sat Jan 1 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: main.c,v 1.20 2001/01/01 18:49:16 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include "graphics.h"
#include "effects.h"
#include "game.h"
#include "map.h"
#include "car.h"
#include "ai.h"
#include "menu.h"
#include "level.h"
#include "score.h"

#ifdef USE_SOUND
#include "sound.h"
#endif 

#define MAPNUM 0
#define FILENAME "default.map"
// If defined, negative delays will make the FRAME_LEN reduce
//#define REARRANGE_FPS
#define FRAME_LEN (long)((long)1000000 / (long)FPS)
#define FUEL_CONSUME 1
#define FUELPTS_SPEEDHACK 100

#define draw_screen() do { \
	update_offset(map, &blue_car); \
	draw_map(map); \
  clean_info_bar(); \
	draw_car(&blue_car); \
	draw_ai(&AI); \
	draw_smoke(); \
	draw_info_bar(); \
	draw_radar(FULL, 0); } while (0)


#include <sys/time.h>
struct timeval st, rt;

// Function declarations
long time_diff(struct timeval *rt);
long delay(long i);
void mapLose();
void mapWin();
int levelWin();
void fuelPts();
int check_player_car();
int initRoutines();
int mainGameLoop();
void goodbyefolks();
void cleanLevel();
char *findDir();

int posX = 0, posY = 0;

extern carStruct blue_car;
extern carStruct red_car;
extern mapStruct *map;
extern aiStruct AI;
int actualFuel;
extern smoke_struct smoke;
extern bool crashed;
extern blend_cache_t blend_cache;
levelStruct level;

bool BYE = false;
bool QUIT = false;
bool USERQUIT = false;
bool CLEAN = true;
bool ENDMAP = false;

int mapNumber = MAPNUM;
int mapResult = 3;
int cnt = 0;

int lives;
char livesString[10];
char mapString[16];
char *hiString;

scoreStruct score;

extern time_t startime;



int main (int argc, char *argv[]) {
//  int cnt = 0, pause = 0;
//	char *levelFile;
	char *dir;
	bool EXITPROGRAM = false;
	int cmd;

	// Initialize the Display
	dbg("Initializing display...");
	if (!init_display("")) {
		hdbg("can't open display");
		return -1;
	}

	dbg("ok");

	// Go to the right directories
	dir = (char *)malloc(256);
	findDir(dir);

	dbg("Working directory is %s", dir);

	dbg("Loading records from ");
	loadRecords("/var/games/xrally.scores");
	dbg("ok");

#ifdef USE_SOUND
	sound_init ();
	dbg("Sound initialized");
	sound_load ("rally.mod");
	sound_switch ();
	sound_select (0);
	sound_volume(64);
	sound_start();
#endif

	atexit (goodbyefolks);

	init_screen();
	initIntro();
	introAnim();
	//
	// Init blend cache
	blend_cache.color.red = blend_cache.color.blue = blend_cache.color.green = 0;
	blend_cache.factor = 60;
	init_blend_cache(&blend_cache);

	while (!EXITPROGRAM) {
		cmd = get_user_option();
		switch(cmd) {
			case PLAY:
				while (initRoutines() == 0);
				break;
			case EXIT:
				EXITPROGRAM = true;
				break;
			case HISCORES:
				displayHiscores(-1, &level, true);
				break;
		}
	}

	goodbyefolks();

	free_blend_cache(&blend_cache);

	return 0;
}

int initRoutines() {
	char *levelFile;
	struct splash_struct splScr;

 	levelFile = chooseLevel();
	if (!levelFile)
		return -1;
	dbg("Level file: %s", levelFile);	

	sprintf(splScr.txt, "Loading %s...", levelFile);
	splScr.n = 9;
	splScr.mode = 0;
	splash(splScr, levelFile);
	
	// Load the level
	if (!loadLevel(&level, levelFile)) {
		hdbg("Can't load level!");
		//initRoutines();
		return -1;
	}
	lives = level.lives;

	dbg("Level loaded");

	sprintf(splScr.txt, "Loading %s tileset...", level.tileset);
	splScr.n = 15 + (level.pics - 5);
	splScr.mode = 1;
	splash(splScr, levelFile);

	// Load the Pixmaps
	if (load_images() < 0) {
		hdbg("Can't load images");
		//initRoutines();
		return -1;
	}
	
	dbg("Images loaded");

	mapNumber = 0;

	if (level.lives <= 0 && (mapNumber = chooseMap(level.number)) < 0) {
		//initRoutines();
		return -1;
	}


	if (lives >= 0)
		sprintf(livesString, "Lives: %d", lives);
	else
		sprintf(livesString, "No Lives!");
		
	// Init score
	sprintf(score.string, "%.*u", SCORE_DIGITS - 1, 0);
	score.value = 0;
	score.bonus = 1;
	score.num_extra = 1;
	score.current_extra = level.extra;

	startime = time(0);

	cleanScr();
	while (mainGameLoop() == 0);
	cleanLevel();

	return 0;
}

int mainGameLoop() {
	long int pause = 0;
	
	ENDMAP = false;
	mapResult = 3;

	if (createMap(&level, mapNumber) < 0) {
		hdbg("Can't load map!");
		return -1;
	}

	dbg("Map %d Loaded", MAPNUM);

	ldbg("Flag num: %d", map->flagNum);

	blue_car.x = toMap(map->startX);
	blue_car.y = toMap(map->startY);
	blue_car.direction = map->startDir;
	blue_car.aiguess = -1;
	blue_car.toGoDir = map->startDir;
	blue_car.vel = map->carVel;
	for (cnt = 0; cnt < SMOKE_NUM; cnt++) {
		smoke.x[cnt] = -1;
		smoke.y[cnt] = -1;
		smoke.left[cnt] = 0;
		smoke.current = 0;
	}
	update_offset(map, &blue_car);
	createAI(&AI, map);
	
	gettimeofday(&rt, NULL);

	// Starting random number generator
	srandom(rt.tv_usec);


#ifdef USE_SOUND
	sound_switch ();
	sound_select(0);
	sound_volume(0);
	sound_start();
#endif

	cleanScr();

  // Get hiscore
  hiString = getRecord(&level);

  sprintf(mapString, "Map: %3d of %3d", mapNumber+1, level.number);
	// Starting sequence
	for (cnt = 0; cnt < 3; cnt++) {
		wait_time(1000000);
		draw_screen();
		draw_light(cnt);
		#ifdef USE_SOUND
		play_sound (SOUND_LIGHT);
		#endif
		graphics_sync();
	}

	QUIT = false;
	USERQUIT = false;

#ifdef USE_SOUND
	sound_select(0);
	sound_volume(level.volume);
#endif

	// REAL GAME LOOP!

	gettimeofday(&st, NULL);
	cnt = 0;
  sprintf(mapString, "Map: %3d of %3d", mapNumber+1, level.number);
	while (!ENDMAP) {
		gettimeofday(&rt, NULL);
		// Playing the game!
		graphics_sync();
		cnt++;
	 	/* Control car routines start */
	 	process_input(check_key());
		// Move player car
		move(&blue_car, map);
		// Check for disasters (AI crash AI or AI crash smoke)
		// Also update the smokes
		carCheck(&AI, map);
		// AI take their reactions
		orderAI(&AI, map);
		moveAI(&AI, map);
		/* Control car routines end */
	
	 	/* Drawing routines  start */
		update_offset(map, &blue_car);
		//cleanScr();
		draw_map(map);
		draw_car(&blue_car);
		draw_ai(&AI);
		draw_smoke();
		clean_info_bar();
		draw_info_bar();
		
		if (map->radarmode != FUELSONAR)
			draw_radar(map->radarmode, map->radarrange);
		else
			draw_radar(SONAR, map->radarrange * (actualFuel<0?0:actualFuel) / map->startFuel);

		// Draw the lights for 1 second)
		if (cnt < FPS)
			draw_light(2);
		/* Drawing routines end */
		
//		check_player_car();


		actualFuel-=FUEL_CONSUME;
		if (actualFuel < 0) {
			blue_car.vel = map->carVel + (actualFuel/map->fuelImpact);
		}
		
		delay(FRAME_LEN - time_diff(&rt));
	}

	pause = time_diff(&st);
	// Ok, ended map!! Now lets do stuff
	ldbg("Ended map... doing stuff");
	dbg("Number of cycles: %d", cnt);
	dbg("Expected frame len: %ld", FRAME_LEN);
	dbg("Real frame len: %ld", pause/cnt);

	if (mapResult == RESULT_WIN) {
		// WON THE GAME
#ifdef USE_SOUND
		sound_volume(0);
#endif
		draw_screen();
		graphics_sync();

		ldbg("Won game");
		sound_switch ();	/* Switch to load in the right slot */
		mapWin();
		// Fool the fps counter
		gettimeofday(&rt, NULL);
	} else if (mapResult == RESULT_LOSE) {
		// LOST THE GAME !!!
#ifdef USE_SOUND
		sound_volume(0);
#endif
		draw_screen();
		graphics_sync();
		wait_time(1000000);
		crashed = false;
		// Fool the fps counter
		gettimeofday(&rt, NULL);
		ldbg("Lost game");
		sound_switch ();	/* Switch to load in the right slot */
		mapLose();
	} else if (mapResult == RESULT_QUIT) {
		// QUITED the game
#ifdef USE_SOUND
		sound_volume(0);
#endif
		crashed = false;
		gettimeofday(&rt, NULL);
	}


	// If its the case, quit the game
	if (level.lives <= 0 && (mapNumber = chooseMap(level.number))<0) {
		return -1;
	}

	// Check if the game is over
	if (level.lives > 0 && (lives < 0 || mapResult == RESULT_QUIT)) {
		sound_stop ();
		sound_switch ();
		dbg ("Add the quit tune here");
		sound_volume (0);
		sound_start ();
		endScreen(&level, score.value, GAME_OVER);
		return -1;
	}
	if (mapNumber >= level.number) {
		endScreen(&level, levelWin(), VICTORY);
		dbg ("Add the victory tune here");
		return -1;
	}
	//mainGameLoop();		/* CM: Recursive function ?!? */
	return 0;
}

void cleanLevel() {
	
	dbg("Ending frame len: %ld", FRAME_LEN);
	dbg("Playing time: %d seconds", cnt / FPS);

	destroyAI();
	free(level.map);
	level.number = 0;
	level.map = NULL;
	fclose(level.file);
	level.file = NULL;

}

void goodbyefolks() {

	static int once = 0;

	if (once)
		return;

	once++;

#ifdef USE_SOUND
	sound_stop();
	dbg("Sound deinitialized");
	sound_deinit();
#endif
	close_display();
	saveRecords("/var/games/xrally.scores");

	return;
}

#ifdef USE_SOUND
void play_sound (int i)
{
	sound_play (i, 60, -1);
}
#endif
		 

long delay(long i)
{
	struct timeval timeout;
	if (i > 0) {
		timeout.tv_usec = i % (long) 1000000;
		timeout.tv_sec = i / (long) 1000000;
		select(0, NULL, NULL, NULL, &timeout);
	}
	return i;
}

long time_diff(struct timeval *ot)
{
 long diff;
 struct timeval nt;

 gettimeofday(&nt, NULL);
 diff = ((long)1000000*(long)(nt.tv_sec-ot->tv_sec))+(long)(nt.tv_usec-ot->tv_usec);
 return diff;
}

void mapLose() {
	if (map->bonus) {
		mapNumber++;
	} else if (level.lives > 0) {
		lives--;
		sprintf(livesString, "Lives: %d", lives);
	}
	score.last = 0;
}

int levelWin() {
	int oldscore = score.value;
	// Add livesbonus to score
	if (level.livesbonus < 0 && (lives-level.lives) > 0)
		add_score(&score, &level, (lives-level.lives)*(-1)*level.livesbonus);
	else if (level.livesbonus > 0)
		add_score(&score, &level, lives * level.livesbonus);
	return oldscore;
}


void mapWin() {
	// Adjust the score
	fuelPts();
	score.last = 0;
	score.timer = 0;
	// Won the game! Next map!
	if (level.lives > 0)
		mapNumber++;
}

void fuelPts() {
	int a = 0;
	if (map->fuelpts == 0)
		return;
	// The actual thing to is:
	// if (map->fuelpts > 0)
	// 	score.value += (actualFuel / map->fuelpts);
	// else
	// 	score.value += abs(actualFuel * map->fuelpts);
	// But lets add some nice effects
	while (actualFuel > 0) {
		a++;
		actualFuel--;
		if (map->fuelpts > 0 && a % map->fuelpts == 0)  // Fuel for points
			add_score(&score, &level, 1);
		else if (map->fuelpts < 0) // Points for fuel
			add_score(&score, &level, -map->fuelpts); // Remenber that fuelpts is negative
		if (a % (map->startFuel/FUELPTS_SPEEDHACK) == 0) {	
			draw_screen();
			graphics_sync();
		}
	}
	draw_screen();
	graphics_sync();
}

char *findDir(char *buf) {
	bool foreign = false;
	int a = 0;
	// Look for maps subdirectory
	foreign = chdir("maps");
	chdir("..");
	// Look for sounds subdirectory
	foreign = chdir("sounds");
	chdir("..");
	// Look for images subdirectory
	foreign = chdir("images");
	chdir("..");
	if (foreign)
		a = chdir(GAMEDIR);
	if (a < 0)
		hdbg("Can't find the game data anywhere! Expect serious problems");
	return getcwd(buf, 256);
}

