/***************************************************************************
                          map.c  -  description
                             -------------------
    begin                : Sat Jan 1 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: map.c,v 1.8 2000/12/23 15:19:33 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "map.h"
#include "score.h"
#include "car.h"
#include "graphics.h"
#include "game.h"
#include "config.h"



extern int actualFuel;


radarStruct radar;

extern scoreStruct score;
extern levelStruct level;
extern int SPRITE_SIZE;
extern int TILE_WIDTH;
extern int TILE_HEIGHT;

// Variable declaration
mapStruct *map;

int createMap(levelStruct *level, int mapNumber) {
	float facX, facY;


	if (mapNumber >= level->number) {
		hdbg("There isn't a map number %d in this file", mapNumber);
		return -1;
	}

	map = &level->map[mapNumber];

	ldbg("Map size:  %d", sizeof(mapStruct));
	ldbg("Data size: %d", sizeof(map->data));


	// Radar:
  // It has a maximum width of 128 and height of 256
	// Check the factors

	facX = RADAR_WIDTH / (float)toMap(map->width);
	facY = RADAR_HEIGHT / (float)toMap(map->height);

	// The fuel should be enough to run 2/3 of the map
	if (map->startFuel < 0)
		map->startFuel = actualFuel = (toMap((map->width)*(map->height)) / STEP * 2) / 3;
	else
		map->startFuel = actualFuel = map->startFuel;
	ldbg("FacX: %f", facX);
	ldbg("FacY: %f", facY);

	if (facX < facY)
		radar.factor = facX;
	else
		radar.factor = facY;

	radar.width = (int)(radar.factor*toMap(1))*map->width;
	radar.height = (int)(radar.factor*toMap(1))*map->height;
	

	return 0;

}

int toTile(int pos) {
	pos+=(SPRITE_SIZE/2);
	pos -= (pos % SPRITE_SIZE);
	return (pos / SPRITE_SIZE);
}

int toMap(int tPos) {
	return (tPos * SPRITE_SIZE);
}

void update_offset(mapStruct *Map, carStruct *Car) {
	int aX, aY;
	// Get the up-right corner
	aX = Car->x - MIDDLE_X;
	aY = Car->y - MIDDLE_Y;
	// OffX and Offy
	Map->offX = aX % SPRITE_SIZE;
	Map->offY = aY % SPRITE_SIZE;
	// tX and tY
	Map->tX = (aX - Map->offX) / SPRITE_SIZE;
	Map->tY = (aY - Map->offY) / SPRITE_SIZE;
}

bool isClean(mapStruct *Map, carStruct *Car, int dir, int lookAhead) {
	int x, y;
	int tX, tY;
	x = Car->x;
	y = Car->y;
	moveXY(&x, &y, Car->direction, lookAhead);
	tX = x / SPRITE_SIZE;
	tY = y / SPRITE_SIZE;
	switch (dir) {
		case UP:
			if (y%SPRITE_SIZE != 0)
				return true; // Can go one more
			else {
				if (x%SPRITE_SIZE == 0)
					return passable(tX, tY-1);
				else
					return (passable(tX, tY-1) && passable(tX+1, tY-1));
			}
			break;
		case DOWN:
			if (y%SPRITE_SIZE != 0)
				return true; // Can go one more
			else {
				if (x%SPRITE_SIZE == 0)
					return passable(tX, tY+1);
				else
					return (passable(tX, tY+1) && passable(tX+1, tY+1));
			}
			break;
		case LEFT:
			if (x%SPRITE_SIZE != 0)
				return true; // Can go one more
			else {
				if (y%SPRITE_SIZE == 0)
					return passable(tX-1, tY);
				else
					return (passable(tX-1, tY) && passable(tX-1, tY+1));
			}
			break;
		case RIGHT:
			if (x%SPRITE_SIZE != 0)
				return true; // Can go one more
			else {
				if (y%SPRITE_SIZE == 0)
					return passable(tX+1, tY);
				else
					return (passable(tX+1, tY) && passable(tX+1, tY+1));
			}
			break;
	}
	return false;
}

bool passable(int x, int y) {
//	if (map->data[ (ty - 1) * map->width + tx ] == BLOCK || map->data[ (ty - 1) * map->width + tx ] >= 9)
	// Check for end of map
	if (x >= map->width || x < 0 || y >= map->height || y < 0)
		return false;
	return !(map->data[y*map->width+x] == BLOCK || map->data[y*map->width+x] >= 9);
}
	/* OLD isClean STUFF
	int tx, ty;
	ty = toTile(Car->y);
	tx = toTile(Car->x);

	// If it is the opposite d, its ok to turn
	if (abs(dir - Car->direction) == 2)
		return true;

	if (lookAhead) {
		switch (Car->direction) {
			case UP:
				ty -= lookAhead;
				break;
			case DOWN:
				ty += lookAhead;
				break;
			case RIGHT:
				tx += lookAhead;
				break;
			case LEFT:
				tx -= lookAhead;
				break;
		}
	}
	switch (dir) {
		// Custom impassable tiles: value >= 11
		case UP:
			// Check if end of map
			if (Car->y < Car->vel)
				return false;
			// Check if wall on top
			if (map->data[ (ty - 1) * map->width + tx ] == BLOCK || map->data[ (ty - 1) * map->width + tx ] >= 9)
				return false;
			break;
		case RIGHT:
			// Check if end of map
			if (Car->x > (toMap(map->width) - Car->vel - SPRITE_SIZE))
				return false;
			// Check if wall at right
			if (map->data[ (ty * map->width) + tx + 1 ] == BLOCK || map->data[ (ty * map->width) + tx + 1 ] >= 9)
				return false;
			break;
		case DOWN:
			// Check if end of map
			if (Car->y > (toMap(map->height) - Car->vel - SPRITE_SIZE))
				return false;
			// Check if wall at down
			if (map->data[ (ty + 1) * map->width + tx ] == BLOCK || map->data[ (ty + 1) * map->width + tx ] >= 9)
				return false;
			break;
		case LEFT:
 			// Check if end of map
			if (Car->x < Car->vel)
				return false;
			// Check if wall at left
			if (map->data[ ty * map->width + tx - 1 ] == BLOCK || map->data[ ty * map->width + tx - 1 ] >= 9)
				return false;
			break;
	}
	return true;
	*/

void getFlag(int x, int y) {
	int a;
	map->data[(y * map->width) + x] = ROAD;
	map->flagLeft--;
	for (a = 0; a < map->flagNum; a++) {
		if (map->flagX[a] == x && map->flagY[a]	== y) {
			ldbg("Got flag %d", a);
			map->flagX[a] = -1;
			map->flagY[a] = -1;
		}
	}
	score.last += FLAG_VALUE;
	score.x = toMap(x);
	score.y = toMap(y);
	score.timer = SCORE_TIMER;
	add_score(&score, &level, score.last * score.bonus);
}

void initMap(mapStruct *map) {
	map->width = -1;
	map->height = -1;
	map->startX = -1;
	map->startY = -1;
	map->red_startX = NULL;
	map->red_startY = NULL;
	map->redNum = -1;
	map->flagNum = -1;
	map->flagX = NULL;
	map->flagY = NULL;	
  map->data = NULL;
	map->startDir = -1;
	map->carVel = STEP;
	map->aiVel = STEP2;
	map->startFuel = -1;
	map->fuelImpact = 1;
	map->fuelpts = 5;
	map->radarmode = ORIGINAL;
	map->radarrange = 480;
	map->enemywait = 0;
}
