/***************************************************************************
                          score.c  -  description
                             -------------------
    begin                : Sat Jan 1 2000
    copyright            : (C) 2000 by Perdig
    email                : perdig@linuxbr.com.br

    $Id: score.c,v 1.6 2001/01/01 18:49:16 perdig Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "config.h"
#include "score.h"
#include "time.h"

#define MAXHISCORE 20

hiScoreStruct hiscore;

#define addMember(member, string) do { \
	hiscoreTemp = member; \
	member = (hiMemberStruct *)malloc(sizeof(hiMemberStruct)); \
	member->name = string; \
	member->child = NULL; \
	member->next = hiscoreTemp; } while (0)
	
bool useGAMEDIR = true;	
time_t startime;

extern int lives;
extern char livesString[10];

hiMemberStruct *lookup(hiMemberStruct *member, char *name) {
	while (member != NULL && strcmp(member->name, name) != 0) {
		member = member->next;
	}
	return member;
}

int numElements(hiMemberStruct *member) {
	int a = 0;
	while (member != NULL) {
		a++;
		member = member->next;
	}
	return a;
}

void loadRecords(char *filename) {
	FILE *fd;
	char buf[256] = "";
	char *temp = NULL;
	hiMemberStruct *hiscoreTemp;
	// Look first for a global hiscore
	temp = (char *)malloc(256);
	getcwd(temp, 256);
	chdir(GAMEDIR);
	fd = fopen(filename, "r");
	if (fd == NULL) {
		useGAMEDIR = false;
		chdir(temp);
		dbg("%s...", temp);
		fd = fopen(filename, "r");
		if (fd == NULL) {
			dbg("can't find a hiscore file. Creating a new one at %s...", temp);
			return;
		}
	} else {
		dbg("%s...", GAMEDIR);
	}
	chdir(temp);
	free(temp);
	temp = NULL;

	while(fgets(buf, 256, fd)) {
		temp = (char *)malloc(strlen(buf-2));
		strcpy(temp, buf+2);
		temp[ strlen(temp)-1 ] = 0;
		switch (*buf) {
			case '*':
				addMember(hiscore.level, temp);
				break;
			case '+':
				addMember(hiscore.level->child, temp);
				break;
			case '-':
				addMember(hiscore.level->child->child, temp);
				break;
		}
	}
}

bool newRecord(levelStruct *level, scoreStruct *score) {
	int record;
	hiMemberStruct *it;
	// If 0 points, don't even care
	if (score->value <= 0)
		return false;

	// Lookup level
	it = lookup(hiscore.level, level->name);

	// Lookup variation
	if (it == NULL) {
		// No level!
		return true;
	}
	it = lookup(it->child, level->sum);
	if (it == NULL) {
		// No var!
		return true;
	}
	it = it->child;
	if (numElements(it) < MAXHISCORE)
		return true;
	else {
		while (it->next != NULL)
			it = it->next;
		sscanf(it->name, "%d", &record);
		return score->value > record;
	}
}

int addRecord(levelStruct *level, scoreStruct *score, char *name) {
	int record = score->value+1;
	hiMemberStruct *itL, *itV, *itR;
	hiMemberStruct *hiscoreTemp;
	char *temp;
	char livStr[16];
	char date[256];
	int pos = 1;
	bool first = false;
	time_t curtime;
	ldbg("Adding record of %d by %s", score->value, name);
	// Create string
	curtime = time(0);
	strftime(date, 256, "%m/%d/%Y", localtime(&curtime));
	temp = (char *)malloc(256);
	if (lives >= 0)
		sprintf(livStr, "%d lives", lives);
	else
		sprintf(livStr, "Died");
	sprintf(temp, "%d %s (%s - %ld min:%ld sec) %s", score->value, name, livStr, (curtime-startime)/60, (curtime-startime)%60, date);
	// Find level
	itL = lookup(hiscore.level, level->name);
	if (itL == NULL) {
		// No level by that name! Create new
		addMember(hiscore.level, level->name);
		itL = hiscore.level;
	}
	// Find sum
	itV = lookup(itL->child, level->sum);
	if (itV == NULL) {
		// No of that sum! Create new
		addMember(itL->child, level->sum);
		itV = itL->child;
	}
	itR = itV->child;
	if (itR == NULL) {
		// No record yet! Create new
		addMember(itV->child, temp);
		return 0;
	}
	sscanf(itR->name, "%d", &record);
	if (record < score->value)
		first = true;
	while (itR->next) {
		sscanf(itR->next->name, "%d", &record);
		if (record < score->value)
			break;
		pos++;
		itR = itR->next;
	}
	// 'itR' now holds a pointer to the right place.
	hiscoreTemp = (hiMemberStruct *)malloc(sizeof(hiMemberStruct));
	hiscoreTemp->next = first ? itR : itR->next;
	hiscoreTemp->child = NULL;
	hiscoreTemp->name = temp;
	if (!first)
		itR->next = hiscoreTemp;
	else
		itV->child = hiscoreTemp;
	// Erases those who are extra
	itR = itV->child;
	while (numElements(itR) > MAXHISCORE) {
		itR = itV->child;
		if (!itR->next)
			break;
		while (itR->next->next != NULL)
			itR = itR->next;
		itR->next = NULL;
	}

	return first ? 0 : pos;
}

char *getRecord(levelStruct *level) {
  hiMemberStruct *itLev;
  hiMemberStruct *itVar;
  unsigned int val;
  char *name = NULL;
  char *retval = NULL;
  for (itLev = hiscore.level; itLev; itLev = itLev->next) {
    if (strcmp(level->name, itLev->name)==0) {
      for (itVar = itLev->child; itVar; itVar = itVar->next) {
        if (strcmp(level->sum, itVar->name)==0) {
          name = itVar->child->name;
          break;
        }
      }
    }
  }
  if (name) {
    retval = (char *)malloc(SCORE_DIGITS*sizeof(char));
    sscanf(name, "%u", &val);
    sprintf(retval, "%.*u", SCORE_DIGITS-1, val);
  }

  return retval;

}

char **listRecords(levelStruct *level) {
	hiMemberStruct *itLev;
	hiMemberStruct *itVar;
	hiMemberStruct *itVal;
	char **values = NULL;
	int a;

	ldbg("Starting to display high scores now!");
	for (itLev = hiscore.level; itLev; itLev = itLev->next) {
		if (strcmp(level->name, itLev->name)==0) {
			for (itVar = itLev->child; itVar; itVar = itVar->next) {
				if (strcmp(level->sum, itVar->name)==0) {
					values = (char **)calloc(numElements(itVar->child)+1, sizeof(char *));
					values[0] = (char *)malloc(sizeof(char));
					*values[0] = (char )numElements(itVar->child);
					for (itVal = itVar->child, a = 1; itVal; itVal = itVal->next, a++) {
						values[a] = itVal->name;
					}
				}
			}
		}
	}
	return values;
}

void saveRecords(char *filename) {
	FILE *fd;
	char *temp = NULL;
	hiMemberStruct *itL, *itV, *itR;
	if (useGAMEDIR) {
		temp = (char *)malloc(256);
		getcwd(temp, 256);
		chdir(GAMEDIR);
	} 
	fd = fopen(filename, "w");
	if (!fd) {
		perror("Can't create new file");
		return;
	}
	if (useGAMEDIR) {
		chdir(temp);
		free(temp);
		temp = NULL;
	}
	// Stores the values in inverse order
	// Also erases everything
	for (itL = hiscore.level; itL; itL = itL->next) {
		// Writes the level
		fprintf(fd, "* %s\n", itL->name);
		for (itV = itL->child; itV; itV = itV->next) {
			fprintf(fd, "+ %s\n", itV->name);
			itR = itV->child;
			// Ok, now saves it
			while (itR != NULL) {
				itR = itV->child;
				if (itR->next == NULL) {
					fprintf(fd, "- %s\n", itR->name);
					break;
				}
				// Puts itR at the end
				while (itR->next->next != NULL)
					itR = itR->next;
				// itR->next is the last one! Write it down
				fprintf(fd, "- %s\n", itR->next->name);
				// Now erases it
				free(itR->next);
				itR->next = NULL;
				// Go to next one
			}
		}
	}
	fclose(fd);
}

void add_score(scoreStruct *score, levelStruct *level, int value) {
	while (value > 0) {
		score->value++;
		value--;
		if (score->value >= score->current_extra && score->current_extra > 0) {
			lives++;
			sprintf(livesString, "Lives: %d", lives);
			score->num_extra++;
			score->current_extra += score->num_extra*level->extra;
			score->timer+=128;
		}
	}
	sprintf(score->string, "%.*u", SCORE_DIGITS - 1, score->value);
}
