/***************************************************************************
                     sound.c  - sound system for xrally 
                             -------------------
    begin                : Sun Jul 23 2000
    copyright            : (C) 2000 by Claudio Matsuoka
    email                : claudio@helllabs.org

    $Id: sound.c,v 1.8 2000/11/25 15:35:19 claudio2 Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include "config.h"
#include "global.h"
#include "soundplayer.h"

/***************************************************************************
 *                                                                         *
 * This player is based on the xmp 2.1 development tree, and has a fairly  *
 * good M.K. MOD replaying support. For more information check the libxmp  *
 * draft at http://xmp.helllabs.org/newxmp.                                *
 *                                                                         *
 ***************************************************************************/

/* Amiga periods */
static int period_amiga[] = {
   /*  0       1       2       3       4       5       6       7   */
    0x1c56, 0x1c22, 0x1bee, 0x1bbb, 0x1b87, 0x1b55, 0x1b22, 0x1af0,  /* B  */
    0x1abf, 0x1a8e, 0x1a5d, 0x1a2c, 0x19fc, 0x19cc, 0x199c, 0x196d,  /* C  */
    0x193e, 0x1910, 0x18e2, 0x18b4, 0x1886, 0x1859, 0x182c, 0x1800,  /* C# */
    0x17d4, 0x17a8, 0x177c, 0x1751, 0x1726, 0x16fb, 0x16d1, 0x16a7,  /* D  */
    0x167d, 0x1654, 0x162b, 0x1602, 0x15d9, 0x15b1, 0x1589, 0x1562,  /* D# */
    0x153a, 0x1513, 0x14ec, 0x14c6, 0x149f, 0x1479, 0x1454, 0x142e,  /* E  */
    0x1409, 0x13e4, 0x13c0, 0x139b, 0x1377, 0x1353, 0x1330, 0x130c,  /* F  */
    0x12e9, 0x12c6, 0x12a4, 0x1282, 0x125f, 0x123e, 0x121c, 0x11fb,  /* F# */
    0x11da, 0x11b9, 0x1198, 0x1178, 0x1157, 0x1137, 0x1118, 0x10f8,  /* G  */
    0x10d9, 0x10ba, 0x109b, 0x107d, 0x105e, 0x1040, 0x1022, 0x1004,  /* G# */
    0x0fe7, 0x0fca, 0x0fad, 0x0f90, 0x0f73, 0x0f57, 0x0f3a, 0x0f1e,  /* A  */
    0x0f02, 0x0ee7, 0x0ecb, 0x0eb0, 0x0e95, 0x0e7a, 0x0e5f, 0x0e45,  /* A# */
    0x0e2b, 0x0e11, 0x0df7, 0x0ddd, 0x0dc3, 0x0daa, 0x0d91, 0x0d78,  /* B  */
};

/* Vibrato/tremolo waveform tables */
static int waveform[4][64] = {
   {   0,  24,  49,  74,  97, 120, 141, 161, 180, 197, 212, 224,
     235, 244, 250, 253, 255, 253, 250, 244, 235, 224, 212, 197,
     180, 161, 141, 120,  97,  74,  49,  24,   0, -24, -49, -74,
     -97,-120,-141,-161,-180,-197,-212,-224,-235,-244,-250,-253,
    -255,-253,-250,-244,-235,-224,-212,-197,-180,-161,-141,-120,
     -97, -74, -49, -24  },	/* Sine */

   {   0,  -8, -16, -24, -32, -40, -48, -56, -64, -72, -80, -88,
     -96,-104,-112,-120,-128,-136,-144,-152,-160,-168,-176,-184,
    -192,-200,-208,-216,-224,-232,-240,-248, 255, 248, 240, 232,
     224, 216, 208, 200, 192, 184, 176, 168, 160, 152, 144, 136,
     128, 120, 112, 104,  96,  88,  80,  72,  64,  56,  48,  40,
      32,  24,  16,   8  },	/* Ramp down */

   { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
     255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
     255, 255, 255, 255, 255, 255, 255, 255,-255,-255,-255,-255,
    -255,-255,-255,-255,-255,-255,-255,-255,-255,-255,-255,-255,
    -255,-255,-255,-255,-255,-255,-255,-255,-255,-255,-255,-255,
    -255,-255,-255,-255  },	/* Square */

   {   0,   8,  16,  24,  32,  40,  48,  56,  64,  72,  80,  88,
      96, 104, 112, 120, 128, 136, 144, 152, 160, 168, 176, 184,
     192, 200, 208, 216, 224, 232, 240, 248,-255,-248,-240,-232,
    -224,-216,-208,-200,-192,-184,-176,-168,-160,-152,-144,-136,
    -128,-120,-112,-104, -96, -88, -80, -72, -64, -56, -48, -40,
     -32, -24, -16,  -8  }	/* Ramp up */
};


static struct mod_priv _priv, *PRIV = &_priv;
static struct mod_state mod[2];
static struct mod_channel channel[NUM_CH + 1];
static char *mod_ptr[2] = { NULL, NULL };
static int slot = 0;

static pthread_t thread;
static pthread_mutex_t frame_mtx = PTHREAD_MUTEX_INITIALIZER;

struct sound_driver *snddev;

int16 *snd_buffer;

static int mix_sound (void)
{
	register int i;
	int c, bsize;

	bsize = BUFFER_SIZE /* * 125 / mod[slot].bpm */;

	memset (snd_buffer, 0, BUFFER_SIZE << 1);

	pthread_mutex_lock (&frame_mtx);

	for (c = 0; c <= NUM_CH; c++) {
		int siz = mod[slot].header->ins[channel[c].sample].size;
		int lsz = mod[slot].header->ins[channel[c].sample].loop_size;
		int lst = mod[slot].header->ins[channel[c].sample].loop_start;
		int per = channel[c].period;
		int vol = channel[c].volume;
		int idx = channel[c].sidx;
		int8 *src = mod[slot].sample[channel[c].sample];
		int16 b;
		//int vib;

		if (!vol || !per || !src)
			continue;

		if (c < NUM_CH)
			vol = mod[slot].volume * vol / 100;

		if (lsz <= 2)
			lsz = 0;
		else if ((lst + lsz) < siz)
			siz = lst + lsz;

		for (i = 0; i < bsize; i++) {
			int hi = idx >> 8;
			int lo = idx & 0xff;

			if (hi >= siz) {
				if (lsz) {
					idx = lst << 8;
					hi = lst;
					lo = 0;
				} else {
					break;
				}
			}

			b = src[hi];
			/* interpolation */
			b += ((src[(hi + 1) % siz] - src[hi]) * lo) >> 8;
			snd_buffer[i] += b * vol;

			idx += 660000 / per;
		}
		channel[c].sidx = idx;
	}
	pthread_mutex_unlock (&frame_mtx);

	return bsize;
}

/* Get period from note and finetune */
static int note_to_period (int n, int b)
{   
	int *t = period_amiga;
	int f = ((b % 100) << 7) / 100; 
    
	if ((n += !((b < 0) && (f += 0x80)) + b / 100) < 0)
		n = 0;
	t += ((n % 12) << 3) + (f >> 4);
	return ((*t << 4) + (*t++ - *t) * (f & 0xf)) >> (n / 12);
}   

/* Get note from period using the Amiga frequency table */
int period_to_note (int p) 
{
	int n, f, *t = period_amiga + MAX_NOTE;

	if (!p)  
		return 0;
	for (n = NOTE_Bb0; p <= (MAX_PERIOD / 2); n += 12, p <<= 1);
	for (; p > *t; t -= 8, n--); 
	for (f = 7; f && (*t > p); t++, f--);       
	return n - (f >> 2);
}   

/* Get pitchbend from base note and period */
int period_to_bend (int p, int n, int f, int g)
{
	int b, *t = period_amiga + MAX_NOTE;

	if (!n) return 0;

	if (p < MIN_PERIOD_A)
		p = MIN_PERIOD_A;

	/* Fixup for panic.s3m  (from Toru Egashira's NSPmod) */
	for (n = NOTE_B0 - 1 - n; p <= (MAX_PERIOD / 2); n += 12, p <<= 1);

	for (; p > *t; t -= 8, n--);
	b = 100 * n + (100 * (*t - p) / (*t - *(t + 8))) +
		((f - 128) % 16) * 100 / 128 * (f < 0 ? -1 : 1);
	return g ? b / 100 * 100 : b;
}

static void event_convert (uint8 *e)
{
	struct mod_event n;

	n.note = period_to_note ((LSN (e[0]) << 8) + e[1]);
	n.ins = ((MSN (e[0]) << 4) | MSN (e[2]));
	n.fxt = LSN (e[2]);
	n.fxp = e[3];

	memcpy (e, &n, 4);
}

static void mod_convert (char *p, int slot)
{
	int i, j, k;
	struct mod_event *e;

	mod[slot].header = (struct mod_header *)p;

	/* convert instrument data */
	for (i = 0; i < 31; i++) {
		B_ENDIAN16 (mod[slot].header->ins[i].size);
		B_ENDIAN16 (mod[slot].header->ins[i].loop_start);
		B_ENDIAN16 (mod[slot].header->ins[i].loop_size);
		mod[slot].header->ins[i].size <<= 1;
		mod[slot].header->ins[i].loop_start <<= 1;
		mod[slot].header->ins[i].loop_size <<= 1;
		mod[slot].header->ins[i].finetune =
			(int8)(mod[slot].header->ins[i].finetune << 4);
	}

	mod[slot].header->pat = 0;
	for (i = 0; i < 128; i++) {
		if (mod[slot].header->order[i] > 0x7f)
			break;
		if (mod[slot].header->order[i] > mod[slot].header->pat)
		mod[slot].header->pat = mod[slot].header->order[i];
	}
	mod[slot].header->pat++;

	mod[slot].pattern = (struct mod_event **)malloc
		(mod[slot].header->pat * sizeof (void *));
	
	/* convert pattern data */
	for (i = 0; i < mod[slot].header->pat; i++) {
		mod[slot].pattern[i] =
			(struct mod_event *)((char *)mod[slot].header +
			sizeof (struct mod_header) + 256 * NUM_CH * i);
		e = mod[slot].pattern[i];
		for (j = 0; j < 64; j++) {
			for (k = 0; k < NUM_CH; k++) {
				event_convert ((char *)e);
				e++;
			}
		}
	}

	mod[slot].sample[0] =
		(char *)mod[slot].pattern[mod[slot].header->pat - 1] +
		256 * NUM_CH;

	for (i = 1; i < 31; i++) {
		mod[slot].sample[i] = mod[slot].sample[i - 1] +
			mod[slot].header->ins[i - 1].size;
	}
}

int mod_play_frame ()
{
	int k;
	struct mod_event *e;

	if (!mod[slot].enabled)
		return 0;

	//cond_signal (XMP_WAIT_FRAME);

	if (!mod[slot].frame) {
		//mod[slot].pattern = PRIV->order[mod[slot].order];
		//mod[slot].nrows = PRIV->rows;
		PRIV->jmp = 0;
		PRIV->pbreak = 0;
		PRIV->pjump = mod[slot].order + 1;
	}

	/*
	 * Process this frame for all sound channels.
	 */

	for (k = 0; k < NUM_CH /*mod[slot].nchannels*/; k++) {
		struct mod_channel *ck = &channel[k];
		//cond_signal (XMP_WAIT_CHN);

		ck->dvolume = ck->dperiod = ck->dpan = 0;

		if (!mod[slot].frame) {
			int i, fxp;

			/*
			 * In the first frame of a row we collect event data,
			 * parse effects, and process "fine" effects. In the
			 * original Protracker playroutine the effect checks
			 * the frame number, this approach should work better.
			 */

			e = mod[slot].pattern[mod[slot].header->order[mod[slot].order]] +
				NUM_CH /*mod[slot].nchannels*/ * mod[slot].row + k;

#if 0
			/* Update channel data */
			ck->event.note = e->note;
			ck->event.ins = e->ins;
			ck->event.vol = 0;	/* FIXME! */
			ck->event.fxp = e->fxp;
			ck->event.fxt = e->fxt;
#endif

			/*
	 	 	 * From the mod replaying notes for PT3.15:
	 		 *
  	 		 *	  Instrument -> None    Same    Valid   Inval
			 * New note		Switch  Play    Play    Cut
			 * New instr. (no note) -       NewVol  NewVol  Cut
			 * Tone portamento	Cont    NewVol  NewVol  Cut
			 *
			 * Play    = Play new note with new default volume
			 * Switch  = Play new note with current volume
			 * NewVol  = Don't play sample, set new default volume
			 * Cont    = Continue playing sample
			 * Cut     = Stop playing sample
			 */

			if (e->ins) {
				ck->volume = mod[slot].header->ins[e->ins - 1].volume;
				ck->fx[finetune].value =
					mod[slot].header->ins[e->ins - 1].finetune;
			}

			if (e->note) {
				ck->note = e->note;

 				if (e->fxt != 0x3 && e->fxt != 0x5) {
					if (e->ins) {
						ck->instr = e->ins - 1;
						ck->sample = ck->instr;
					}

					ck->sidx = 0;
					ck->period = ck->fx[portamento].target =
						note_to_period (e->note,
						ck->fx[finetune].value);
				}
			}

			/*
			 * effects
			 */

			for (i = 0; i < fx_last; i++)
				fx_disable (ck->fx[i]);

			switch (e->fxt) {
			case 0x0:	/* arpeggio */
				if (!e->fxp)
					break;
				fx_enable (ck->fx[arpeggio]);
				ck->fx[arpeggio].data[0] = 0;
				ck->fx[arpeggio].data[1] = MSN(e->fxp);
				ck->fx[arpeggio].data[2] = LSN(e->fxp);
				ck->fx[arpeggio].value = 0;;
				break;
			case 0x1:	/* pitchbend up */
				fx_enable (ck->fx[pitchbend]);
				ck->fx[pitchbend].value = -e->fxp;
				break;
			case 0x2:	/* pitchbend down */
				fx_enable (ck->fx[pitchbend]);
				ck->fx[pitchbend].value = e->fxp;
				break;
			case 0x3:	/* tone portamento */
				fx_enable (ck->fx[portamento]);
				if (e->note) {
					ck->fx[portamento].target =
						note_to_period (e->note,
						ck->fx[finetune].value);
				}
				if (e->fxp)
					ck->fx[portamento].value = e->fxp;
				break;
			case 0x4:	/* vibrato */
				fx_enable (ck->fx[vibrato]);
				if (e->fxp) {
					ck->fx[vibrato].vdepth = LSN (e->fxp);
					ck->fx[vibrato].vspeed = MSN (e->fxp);
				}
				break;
			case 0x5:	/* portamento + vslide */
				fx_enable (ck->fx[portamento]);
				fx_enable (ck->fx[vslide]);
				if (e->note) {
					ck->fx[portamento].target =
						note_to_period (e->note,
						ck->fx[finetune].value);
				}
				ck->fx[vslide].value = MSN(e->fxp)-LSN(e->fxp);
				break;
			case 0x6:	/* vibrato + vslide */
				fx_enable (ck->fx[vibrato]);
				fx_enable (ck->fx[vslide]);
				ck->fx[vslide].value = MSN(e->fxp)-LSN(e->fxp);
				break;
			case 0x7:	/* tremolo */
				fx_enable (ck->fx[tremolo]);
				if (e->fxp) {
					ck->fx[tremolo].vdepth = LSN (e->fxp);
					ck->fx[tremolo].vspeed = MSN (e->fxp);
				}
				break;
			case 0x8:	/* pan setting -- not standard */
				/* not implemented */
				break;
			case 0x9:
				ck->sidx = e->fxp << 16;
				break;
			case 0xa:	/* volume slide */
				fx_enable (ck->fx[vslide]);
				ck->fx[vslide].value = MSN(e->fxp)-LSN(e->fxp);
				break;
			case 0xb:	/* pattern jump */
				PRIV->jmp = 1;
				PRIV->pjump = e->fxp;
				break;
			case 0xc:	/* volume set */
				ck->volume = e->fxp;
				break;
			case 0xd:	/* pattern break */
				PRIV->jmp = 1;
				PRIV->pbreak = (e->fxp >> 4) * 10 +
					(e->fxp & 0x0f);
				break;
			case 0xe:	/* extended effect */
				fx_enable (ck->fx[extended]);
				break;
			case 0xf:
				if (e->fxp == 0)
					break;
				if (e->fxp > 0x20)
					;//mod[slot].bpm = e->fxp;
				else
					mod[slot].speed = e->fxp;
				break;
			}

			if (!fx_test (ck->fx[extended]))
				goto skip_extended;

			/*
			 * extended effects
			 */

			fxp = e->fxp & 0xf;

			switch (e->fxp >> 4) {
			case 0x0:	/* filter on/off */
				/* not implemented */
				break;
			case 0x1:	/* fine portamento up */
				fx_enable (ck->fx[fineporta]);
				ck->fx[fineporta].value = -fxp;
				break;
			case 0x2:	/* fine portamento down */
				fx_enable (ck->fx[fineporta]);
				ck->fx[fineporta].value = fxp;
				break;
			case 0x3:	/* glissando */
				break;
			case 0x4:	/* vibrato waveform */
				PRIV->vibrato_wf = fxp;
				break;
			case 0x5:	/* set finetune */
				ck->fx[finetune].value = (uint8)(fxp << 4);
				break;
			case 0x6:	/* pattern loop */
				if (fxp == 0) {
					ck->fx[loop].data[0] = mod[slot].row;
					break;
				}

				if (ck->fx[loop].value == 0) {
					ck->fx[loop].value = fxp;
				} else {
					ck->fx[loop].value--;
					if (ck->fx[loop].value == 0)
						break;
				}

				PRIV->jmp = 1;
				PRIV->pjump = mod[slot].order;
				PRIV->pbreak = ck->fx[loop].data[0];
				break;
			case 0x7:	/* tremolo waveform */
				PRIV->tremolo_wf = fxp;
				break;
			case 0x8:	/* */
				break;
			case 0x9:	/* note retrig */
				fx_enable (ck->fx[retrig]);
				ck->fx[retrig].value = fxp;
				ck->fx[retrig].data[0] = 1;
				ck->fx[retrig].data[1] = 1;
				ck->fx[retrig].data[2] = ck->volume;
				break;
			case 0xa:	/* fine volslide up */
				fx_enable (ck->fx[finevol]);
				ck->fx[finevol].value = fxp;
				break;
			case 0xb:	/* fine volslide down */
				fx_enable (ck->fx[finevol]);
				ck->fx[finevol].value = -fxp;
				break;
			case 0xc:	/* note cut */
				fx_enable (ck->fx[retrig]);
				ck->fx[retrig].value = fxp;
				ck->fx[retrig].data[0] = 1;
				ck->fx[retrig].data[1] = 0;
				ck->fx[retrig].data[2] = ck->volume;
				break;
			case 0xd:	/* note delay */
				fx_enable (ck->fx[retrig]);
				ck->fx[retrig].value = fxp;
				ck->fx[retrig].data[0] = 0;
				ck->fx[retrig].data[1] = 1;
				ck->fx[retrig].data[2] = ck->volume;
				break;
			case 0xe:	/* pattern delay */
				ck->fx[pdelay].value = fxp;
				break;
			case 0xf:	/* not defined */
				break;
			}
	
			/*
			 * "fine" effect processing
			 */

			if (fx_test (ck->fx[finevol]))
				ck->volume += ck->fx[finevol].value;

			if (fx_test (ck->fx[fineporta]))
				ck->period += ck->fx[fineporta].value;

			/*
			 * Retrig (cut, delay) processing
			 */
			if (fx_test (ck->fx[retrig])) {
				ck->volume = ck->fx[retrig].data[0] *
					ck->fx[retrig].data[2];
			}
skip_extended:
			;
		} else {
			/*
			 * In the rest of the frames we slide volume, pitch,
			 * etc. as set in the first frame of the row.
			 */

			if (fx_test (ck->fx[vslide]))
				ck->volume += ck->fx[vslide].value;

			if (fx_test (ck->fx[pitchbend]))
				ck->period += ck->fx[pitchbend].value;

			if (fx_test (ck->fx[portamento])) {
				if (abs (ck->fx[portamento].target - ck->period)
					< abs (ck->fx[portamento].value)) {
					ck->period = ck->fx[portamento].target;
					fx_disable (ck->fx[portamento]);
				} else {
					ck->period += (ck->fx[portamento].target
						> ck->period ? 1 : -1) *
						ck->fx[portamento].value;
				}
			}
		}

		/*
		 * Arpeggio, vibrato and tremolo are processed in all frames.
		 * We also handle retrig events here.
		 */

		if (fx_test (ck->fx[arpeggio])) {
			ck->period = note_to_period (ck->note +
				ck->fx[arpeggio].data[ck->fx[arpeggio].value++],
				ck->fx[finetune].value);
			ck->fx[arpeggio].value %= 3;
		}

		if (fx_test (ck->fx[vibrato])) {
			ck->fx[vibrato].vphase %= 64;
			ck->dperiod = (waveform[PRIV->vibrato_wf]
				[ck->fx[vibrato].vphase] *
				ck->fx[vibrato].vdepth) >> 9;
			ck->fx[vibrato].vphase += ck->fx[vibrato].vspeed;
		}

		if (fx_test (ck->fx[tremolo])) {
			ck->fx[tremolo].vphase %= 64;
			ck->dvolume = (waveform[PRIV->tremolo_wf]
				[ck->fx[tremolo].vphase] *
				ck->fx[tremolo].vdepth) >> 9;
			ck->fx[tremolo].vphase += ck->fx[tremolo].vspeed;
		}

		if (fx_test (ck->fx[retrig])) {
			if (ck->fx[retrig].value == 0) {
				ck->sidx = 0;
				ck->volume = ck->fx[retrig].data[1] *
					ck->fx[retrig].data[2];
				fx_disable (ck->fx[retrig]);
			} else {
				ck->fx[retrig].value--;
			}
		}

		/*
		 * Volume and period clipping
		 */

		if (ck->volume > 64)
			ck->volume = 64;
		if (ck->volume < 0)
			ck->volume = 0;

		if (ck->period < MIN_PERIOD_A)
			ck->period = MIN_PERIOD_A;
		else if (ck->period > MAX_PERIOD_A)
			ck->period = MAX_PERIOD_A;
	}

	/*
	 * Update xmp variables
	 */

	mod[slot].frame++;

	if (mod[slot].frame < mod[slot].speed)
		return 0;

	mod[slot].frame = 0;
	mod[slot].row++;

	//cond_signal (XMP_WAIT_ROW);

	if (!PRIV->jmp && mod[slot].row < 64)
		return 0;

	mod[slot].row = PRIV->pbreak;
	mod[slot].order = PRIV->pjump;

	//cond_signal (XMP_WAIT_ORD);

	if (mod[slot].order < mod[slot].header->len)
		return 0;

	//mod[slot].loop++;
	mod[slot].order = 0;
	//cond_signal (XMP_WAIT_LOOP);

	return 0;
}

static void *play_thread (void *arg)
{
	while (mod[slot].playing) {
		mod_play_frame ();
		mix_sound ();
		snddev->dump_buffer ();
	}

	pthread_exit (NULL);
	return NULL;
}


void sound_switch ()
{
	slot ^= 1;
	dbg ("Switching to sound slot %d", slot);
}

void sound_select (int s)
{
	dbg ("Selecting sound %d (slot %d)", s, slot);
	
	mod[slot].order = s;
	mod[slot].row = 0;
	mod[slot].frame = 0;
	mod[slot].playing = 1;
}

void sound_start ()
{
	dbg ("Starting sound (slot %d)", slot);

	if (!mod[slot].enabled)
		return;

	if (mod_ptr[slot] == NULL)
		return;

	dbg ("Creating sound thread");
	pthread_create (&thread, NULL, play_thread, NULL);
	pthread_detach (thread);
}

void sound_stop ()
{
	dbg ("Stopping sound (slot %d)", slot);
	mod[slot].playing = 0;
}

int sound_load (char *m)
{
	FILE *f;
	struct stat st;
	int slot2 = slot ^ 1;	/* Work on the inactive slot */

	dbg ("Loading %s in slot %d", m, slot2);

	mod[slot2].enabled = 0;
	if (chdir("sounds") < 0) {
		fprintf(stderr, "can't go to sounds directory from %s...",
			getcwd(NULL, -1));
		return -1;
	}

	f = fopen (m, "r");
	if (f == NULL) {
		chdir("..");
		fprintf(stderr, "can't load file...");
		return -1;
	}

	fstat (fileno (f), &st);

	if (mod_ptr[slot2] != NULL)
		free (mod_ptr[slot2]);
	mod_ptr[slot2] = malloc (st.st_size);
	fread (mod_ptr[slot2], 1, st.st_size, f);

	mod_convert (mod_ptr[slot2], slot2);
	mod[slot2].speed = 6;
	mod[slot2].enabled = 1;

	chdir("..");
	return 0;
}

void sound_play (int s, int n, int v)
{
	pthread_mutex_lock (&frame_mtx);

	channel[NUM_CH].note = n;
	channel[NUM_CH].period = note_to_period (n, 0);
	channel[NUM_CH].volume = v < 0 ? (mod[slot].header ?
		mod[slot].header->ins[s].volume : 0) : v;
	channel[NUM_CH].sample = s;
	channel[NUM_CH].sidx = 0;

	pthread_mutex_unlock (&frame_mtx);
}

int sound_init ()
{
	dbg ("Initializing sound");
	snd_buffer = malloc (BUFFER_SIZE * sizeof (int16));

	if (sound_oss.init () == 0) {
		snddev = &sound_oss;
		return 0;
	}
	snddev = &sound_dummy;
	return 0;
}

int sound_playing ()
{
	return mod[slot].playing; 
}

void sound_deinit ()
{
	dbg ("Deinitializing sound");
	sound_unload ();
	snddev->deinit ();
	free (snd_buffer);
}

void sound_volume (int i)
{
	mod[slot].volume = i;
}

void sound_unload ()
{
	dbg ("Unloading mod in slot %d", slot);
	mod[slot].enabled = 0;
	sound_stop ();
	pthread_join (thread, NULL);
	free (mod_ptr[slot]);
	mod_ptr[slot] = NULL;
}

void sound_changevol (int i)
{
	mod[slot].volume += i;
	if (mod[slot].volume < 0)
		mod[slot].volume = 0;
	if (mod[slot].volume > 64)
		mod[slot].volume = 64;
}
