/***************************************************************************
                   soundplayer.h  -  sound player internals
                             -------------------
    begin                : Sun Jul 23 2000
    copyright            : (C) 2000 by Claudio Matsuoka
    email                : claudio@helllabs.org

    $Id: soundplayer.h,v 1.1 2000/11/22 22:43:25 claudio2 Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _XRALLY_SOUNDPLAYER_H
#define _XRALLY_SOUNDPLAYER_H

#include <stdio.h>
#include "sound.h"

#ifdef __GNUC__
#define PACKED __attribute((packed))
#endif

#define NUM_CH		4
#define BUFFER_SIZE	440

#define MSN(x)		(((x)&0xf0)>>4) 
#define LSN(x)		((x)&0x0f)

/* Endianism fixup */
#define FIX_ENDIANISM_16(x)	(x=((((x)&0xff00)>>8)|(((x)&0xff)<<8)))

#ifdef WORDS_BIGENDIAN
#define L_ENDIAN16(x)	FIX_ENDIANISM_16(x)
#define B_ENDIAN16(x)	(x=x)
#else
#define L_ENDIAN16(x)	(x=x)
#define B_ENDIAN16(x)	FIX_ENDIANISM_16(x)
#endif

/* Macros for period conversion */
#define NOTE_B0		11
#define NOTE_Bb0	(NOTE_B0+1)
#define MAX_NOTE	(NOTE_B0*8)
#define MAX_PERIOD	0x1c56
#define MIN_PERIOD_A	0x0008
#define MAX_PERIOD_A	0x1abf
#define MIN_PERIOD_L	0x0000
#define MAX_PERIOD_L	0x1e00

/* Amiga limits */
#define AMIGA_LIMIT_UPPER 108
#define AMIGA_LIMIT_LOWER 907

typedef signed char int8;
typedef unsigned char uint8;
typedef signed short int16;
typedef unsigned short uint16;

/* Player structures */

struct mod_state {
	int enabled;
	int volume;
	int playing;
	int speed;
	int order;
	int row;
	int frame;

	struct mod_header *header;
	char *sample[31];
	struct mod_event **pattern;
};

struct mod_priv {
	struct mod_header *header;
	char *sample[31];
	uint8 *order;
	struct mod_event **pattern;
	int vibrato_wf;
	int tremolo_wf;
	int rows;
	int jmp;
	int pbreak;
	int pjump;
};

typedef struct xmp_effect {
	int active;
	int value;
	int data[4];
} xmp_fx;

#define vspeed data[0]
#define vdepth data[1]
#define vphase data[2]
#define target data[0]

#define fx_enable(x) do { x.active = 1; } while (0)
#define fx_disable(x) do { x.active = 0; } while (0)
#define fx_test(x) (x.active != 0)

/* Sound driver stuff */

extern int16 *snd_buffer;
extern struct sound_driver *snddev;


/* MOD stuff */

enum effects {
	arpeggio = 0,
	extended,
	fineporta,
	finetune,
	finevol,
	loop,
	pitchbend,
	pdelay,
	portamento,
	retrig,
	tremolo,
	vibrato,
	vslide,

	fx_last
};
 
struct mod_channel {
	int note;
	int instr;
	int volume;
	int period;
	int pan;
	int dvolume;
	int dperiod;
	int dpan;
	int sample;
	int sidx;
	int flags;
	xmp_fx fx[fx_last];
};

struct mod_instrument {
	uint8 name[22];		/* Instrument name */
	uint16 size;		/* Sample length in 16-bit words */
	int8 finetune;		/* Finetune (signed nibble) */
	int8 volume;		/* Linear playback volume */
	uint16 loop_start;	/* Loop start in 16-bit words */
	uint16 loop_size;	/* Loop length in 16-bit words */
} PACKED;

struct mod_header {
	uint8 name[20];
	struct mod_instrument ins[31];
	uint8 len;
	uint8 pat;		/* We'll use this as number of patterns */
	uint8 order[128];
	uint8 magic[4];
} PACKED;

struct mod_event {
	uint8 note;
	uint8 ins;
	uint8 fxt;
	uint8 fxp;
} PACKED;


#endif	/* _XRALLY_SOUNDPLAYER_H */
